### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''Area of interest.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__date__    = '$Date: 2009-01-03 12:54:43 +0100 (sab, 03 gen 2009) $'
__version__ = '$Revision: 700 $'

import os

import gtk

import dialogs
import gladewrap
from utils import remove_quote, get_gladefile

name_to_label = {
    'coorsys':      'Coordinate System',
    'topleft':      'Top Left Corner',
    'bottomright':  'Bottom Right Corner',
    'topright':     'Top Right Corner',
    'bottomleft':   'Bottom Left Corner',
    'centre':       'Centre',
    'sizeunit':     'Size Unit',
    'size':         'Size',
    'vertexnum':    'Number of Vertex',
    'vertexlist':   'Vertex',
}

class ItemsEditor(gladewrap.GladeObject):
    '''Graphical items editor.

    This class provides a GUI with base functionalities for items editing.
    It is implemented using gtk and glade.
    '''
    root = 'items_editor_hbox'

    def __init__(self, model=None, columns=None):
        gladefile = get_gladefile('bestgui.glade')
        super(ItemsEditor, self).__init__(gladefile, self.root)
        self.view = self.items_editor_treeview

        if model is None and not columns:
            model   = gtk.ListStore(str)
            def on_cell_edited(cell, path, newtest, model, column):
                if newtest:
                    model[path][column] = newtest
            cellrenderer = gtk.CellRendererText()
            cellrenderer.set_property('editable', True)
            cellrenderer.connect('edited', on_cell_edited, model, 0)
            columns = (gtk.TreeViewColumn('Label', cellrenderer, text=0),)

        self.view.set_model(model)
        for col in columns:
            self.view.append_column(col)

        self.remove_item_button.set_sensitive(False)
        self.edit_item_button.set_sensitive(False)
        self.reset_items_button.set_sensitive(False)

    def get_values(self, vlist=None):
        model = self.view.get_model()
        if vlist is None:
            vlist = []
        for row in model:
            itemlist = []
            for elem in row:
                itemlist.append(elem)
            vlist.append(tuple(itemlist))
        return vlist

    def set_values(self, vlist):
        model = self.view.get_model()
        self.on_clear_all_items()
        for item in vlist:
            model.append(item)

    def on_add_item(self, widget=None, data=None):
        model = self.view.get_model()
        modeliter = model.append()
        self.remove_item_button.set_sensitive(True)
        return(modeliter)

    def on_remove_item(self, widget=None, data=None):
        model, modeliter = self.view.get_selection().get_selected()
        if modeliter:
            model.remove(modeliter)

        # use this if multiple selection is enabled
        #~ model, paths = self.view.get_selection().get_selected_rows()
        #~ for path in paths:
            #~ modeliter = model.get_iter(path)
            #~ model.remove(modeliter)

        if len(model) <= 0:
            # disable remove_item_button
            self.remove_item_button.set_sensitive(False)

    def on_edit_item(self, widget=None, data=None):
        pass

    def on_edit_item_event(self, widget, event, data=None):
        return(self.on_edit_item(widget, data))

    def on_move_up_item(self, widget=None, data=None):
        model, modeliter = self.view.get_selection().get_selected()
        if modeliter:
            indx = model.get_path(modeliter)[0]
            if indx == 0:
                return
            previter = model.get_iter(indx-1)
            model.swap(modeliter, previter)

    def on_move_down_item(self, widget=None, data=None):
        model, modeliter = self.view.get_selection().get_selected()
        if modeliter:
            indx = model.get_path(modeliter)[0]
            if indx == len(model)-1:
                return
            nextiter = model.get_iter(indx+1)
            model.swap(modeliter, nextiter)

    def on_clear_all_items(self, widget=None, data=None):
        model = self.view.get_model()
        model.clear()

    def on_reset_items(self, widget=None, data=None):
        pass

class VertexEditor(ItemsEditor):
    '''Graphical vertex editor for AoI selection.'''
    coorsys_labels_map = {
        'ROWCOL': ('Row', 'Column'),
        'LATLON': ('Latitude', 'Longitude'),
    }

    validation_func_map = {
        'ROWCOL': lambda x : int(float(x)),
        'LATLON': float,
        None: str,
    }

    coorsys_index_map = {
        'ROWCOL': 0,
        'LATLON': 1,
        0: 'ROWCOL',
        1: 'LATLON',
    }

    root = 'vertex_editor_vbox'

    def __init__(self, coorsys='ROWCOL', vertex='', editable_cells=False):
        model = gtk.ListStore(int, str, str)
        labels = ['N.'] + list(self.coorsys_labels_map[coorsys])
        columns = []

        def rowindex_datafunc(column, cellrenderer, model, modeliter,
                              data=None):
            path = model.get_path(modeliter)
            cellrenderer.set_property('text', path[0]+1)

        for indx, label in enumerate(labels):
            cellrenderer = gtk.CellRendererText()
            cellrenderer.connect('edited', self.on_cell_edited, model, indx)
            col = gtk.TreeViewColumn(label, cellrenderer, text=indx)
            col.set_resizable(True)
            if indx == 0:
                col.set_min_width(50)
                col.set_cell_data_func(cellrenderer, rowindex_datafunc)
            else:
                col.set_min_width(150)
            columns.append(col)

        super(VertexEditor, self).__init__(model, columns)
        self.edit_item_button.set_sensitive(True)
        self.reset_items_button.set_sensitive(True)
        self.set_editable_cells(editable_cells)
        self.backup = {}
        self.defaultvalues = ( ('Coordinate System', coorsys),
                               ('Vertex', vertex) )
        self.set_values(self.defaultvalues)
        self.coorsys_combobox.set_active(self.coorsys_index_map[coorsys])

    def get_coorsys(self):
        columnview = self.view.get_column(1)
        title = columnview.get_title()
        for coorsys in self.coorsys_labels_map.keys():
            if title in self.coorsys_labels_map[coorsys]:
                return coorsys
        return None

    def set_coorsys(self, coorsys):
        current_coorsys = self.get_coorsys()
        if current_coorsys == coorsys:
            return
        else:
            self.backup[current_coorsys] = self.get_values()
        self.on_clear_all_items()
        for indx, title in enumerate(self.coorsys_labels_map[coorsys]):
            columnview = self.view.get_column(indx+1)
            columnview.set_title(title)
        values = self.backup.get(coorsys)
        if values:
            values = [(key.lower(), val) for key, val in values
                            if key.lower() != name_to_label['coorsys'].lower()]
            self.set_values(values)

    def set_coorsys_enabled(self, enabled):
        self.coorsys_combobox.set_sensitive(enabled)

    def get_values(self, vlist=None):
        model = self.view.get_model()
        vertex = []
        for row in model:
            vertex.append(row[1])
            vertex.append(row[2])

        if vlist is None:
            vlist = []
        if self.coorsys_combobox.get_property('sensitive'):
            vlist.append(
                (name_to_label['coorsys'], '"%s"' % self.get_coorsys()) )
        vlist.append( (name_to_label['vertexnum'], str(len(model))) )
        vlist.append( (name_to_label['vertexlist'],     ','.join(vertex)) )
        return vlist

    def set_values(self, vlist):
        '''Set values specified in vlist. vlist can be both a sequence of
        couples (key, value) and a dictionary like object.'''

        dict_ = dict(vlist)

        # Set the coordinate system
        key = name_to_label['coorsys'].lower()
        value = dict_.get(key)
        if value:
            self.set_coorsys(remove_quote(value))

        # Set vertex
        key = name_to_label['vertexlist'].lower()
        value = dict_.get(key)
        if value:
            oldvalues = self.get_values()
            coorsys = self.get_coorsys()
            checkfunc = self.validation_func_map[coorsys]

            model = self.view.get_model()
            vertex = [val.strip() for val in value.split(',')]
            try:
                for c1, c2 in zip(vertex[0::2], vertex[1::2]):
                    c1 = checkfunc(c1)
                    c2 = checkfunc(c2)
                    model.append( (len(model), str(c1), str(c2)) )
            except ValueError, e:
                msg = 'Bad formatted input file.\n\n' + str(e)
                dialogs.run_error_dialog(msg)
                self.set_values(oldvalues)

    def set_values_from_cfg(self, cfg, section):
        values = [(key, value) for key, value in cfg.items(section)]
        self.set_values(values)

    def get_editable_cells(self):
        column = self.view.get_column(1)
        cellrenderer = column.get_cell_renderers()[0]
        return(cellrenderer.get_property('editable'))

    def set_editable_cells(self, editable=True):
        for column in self.view.get_columns()[1:3]:
            renderers = column.get_cell_renderers()
            #cellrenderer[0].editable = editable # why doesn' work?
            renderers[0].set_property('editable', editable)

    def on_cell_edited(self, cell, path, newtest, model, column):
        newtest = newtest.strip()
        if not newtest:
            return
        coorsys = self.get_coorsys()
        func = self.validation_func_map[coorsys]
        try:
            model[path][column] = func(newtest)
        except ValueError, e:
            dialogs.run_error_dialog(str(e))

    def on_add_item(self, widget=None, data=None):
        coorsys = self.get_coorsys()
        dialog = dialogs.EditCoorDialog(parent=self.widget.get_toplevel(),
                                        coorsys=coorsys)
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            coor = dialog.get_coor()
            model = self.view.get_model()
            modeliter = model.append((len(model), str(coor[0]), str(coor[1])))
            self.remove_item_button.set_sensitive(True)
        else:
            modeliter = None
        dialog.destroy()
        return(modeliter)

    def on_edit_item(self, widget=None, data=None):
        model, modeliter = self.view.get_selection().get_selected()
        if not modeliter:
            return
        coor1 = model.get_value(modeliter, 1)
        coor2 = model.get_value(modeliter, 2)
        coorsys = self.get_coorsys()
        dialog = dialogs.EditCoorDialog(parent=self.widget.get_toplevel(),
                                        coorsys=coorsys, coor=(coor1, coor2))
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            coor = dialog.get_coor()
            model.set_value(modeliter, 1, coor[0])
            model.set_value(modeliter, 2, coor[1])
        dialog.destroy()

    def on_reset_items(self, widget=None, data=None):
        self.on_clear_all_items()
        self.set_values(self.defaultvalues)
        if len(self.view.get_model()) == 0:
            self.remove_item_button.set_sensitive(False)

    def on_coorsys_changed(self, widget, data=None):
        indx = self.coorsys_combobox.get_active()
        self.set_coorsys(self.coorsys_index_map[indx])

class VertexEditorDialog(gladewrap.GladeDialog, VertexEditor):
    '''Graphical vertex editor dialog for AoI selection.'''

    root = 'vertexEditorDialog'

    patterns = (
        ('*.txt *.dat', 'ASCII files'),
        ('*.txt', 'Text files'),
        ('*.dat', 'ASCII data files'),
        ('*', 'All Files'),
    )

    def __init__(self, coorsys='ROWCOL', vertex='', editable_cells=False):
        VertexEditor.__init__(self, coorsys, vertex, editable_cells)
        # @TODO: __init__ method from base class 'gladewrap.GladeDialog'
        #        is not called
        self.filename = None

    def load(self, filename):
        if not filename:
            return
        try:
            fileobj = open(filename)
            try:
                coor = ', '.join([line.rstrip() for line in fileobj])
            finally:
                fileobj.close()
            if coor:
                values = {}
                key = name_to_label['vertexlist'].lower()
                values[key] = coor
                self.set_values(values)
            self.filename_entry.set_text(os.path.basename(filename))
            self.filename = filename
        except (IOError, OSError), e:
            msg = 'Unable to open "%s".\n\n%s' % (filename, str(e))
            dialogs.run_error_dialog(msg)

    def on_open(self, widget=None, data=None):
        filedialog = gtk.FileChooserDialog(
                        title='Open',
                        parent=self.widget,
                        buttons=(gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                 gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        filedialog.set_default_response(gtk.RESPONSE_OK)
        dialogs.add_filechooser_filters(filedialog, self.patterns)
        if self.filename:
            filedialog.set_filename(self.filename)

        response = filedialog.run()
        if response == gtk.RESPONSE_OK:
            filename = filedialog.get_filename()
            if filename:
                self.load(filename)
        filedialog.destroy()

    def on_save(self, widget=None, data=None):
        filename = self.filename
        if filename:
            try:
                fileobj = open(filename, 'w')
                try:
                    values = dict(self.get_values())
                    values = values[name_to_label['vertexlist']]
                    values = values.split(',')
                    values = zip(values[0::2], values[1::2])
                    lines = [', '.join((row, col)) for row, col in values]
                    fileobj.write('\n'.join(lines))
                    return True
                finally:
                    fileobj.close()
            except (IOError, OSError), e:
                dialogs.run_error_dialog(e)
                return False
        else:
            return self.on_saveas(widget, data)

    def on_saveas(self, widget=None, data=None):
        filedialog = gtk.FileChooserDialog(
                            title   = 'Save As',
                            parent  = self.widget,
                            action  = gtk.FILE_CHOOSER_ACTION_SAVE,
                            buttons = (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        filedialog.set_select_multiple(False)
        filedialog.set_default_response(gtk.RESPONSE_OK)
        dialogs.add_filechooser_filters(filedialog, self.patterns)

        # @TODO: use workdr
        filename = self.filename
        if os.path.isabs(filename):
            filedialog.set_filename(filename)
        else:
            filedialog.set_current_name(os.path.basename(filename))

        filename = dialogs.run_save_dialog(filedialog)
        filedialog.destroy()
        if filename:
            self.filename = filename
            self.filename_entry.set_text(os.path.basename(filename))
            return self.on_save(widget, data)


class AoIForm(gladewrap.GladeObject):
    '''Area of Interest (AoI) form'''
    M_TOPLEFTBOTTOMRIGHT = 0
    M_TOPRIGHTBOTTOLEFT  = 1
    M_CENTRESIZE         = 2
    M_VERTEX             = 3

    mode_labels_map = {
        M_TOPLEFTBOTTOMRIGHT: 'Top Left and Bottom Right Corners',
        M_TOPRIGHTBOTTOLEFT:  'Top Right and Bottom Left Corners',
        M_CENTRESIZE:         'Centre and Size',
        M_VERTEX:             'Vertex List',
        'Top Left and Bottom Right Corners': M_TOPLEFTBOTTOMRIGHT,
        'Top Right and Bottom Left Corners': M_TOPRIGHTBOTTOLEFT,
        'Centre and Size':    M_CENTRESIZE,
        'Vertex List':        M_VERTEX,
    }

    units_index_map = {
        'PIXEL': 0,
        'KM':    1,
        0:       'PIXEL',
        1:       'KM',
    }

    coorsys_index_map = {
        'ROWCOL': 0,
        'LATLON': 1,
        0:        'ROWCOL',
        1:        'LATLON',
    }

    coorsys_labels_map = {
        'ROWCOL': ('Row', 'Column'),
        'LATLON': ('Latitude', 'Longitude')
    }

    root = 'main_AoI_vbox'

    def __init__(self, mode=M_TOPLEFTBOTTOMRIGHT, coorsys='ROWCOL'):
        gladefile = get_gladefile('bestgui.glade')
        super(AoIForm, self).__init__(gladefile, self.root)

        self.vertexeditor = VertexEditor(coorsys=coorsys)
        self.notebook.append_page(self.vertexeditor.widget,
                                  gtk.Label('Vertex'))

        self.format_combobox.set_active(mode)
        self.corners_coorsys_combobox.set_active(
                                            self.coorsys_index_map[coorsys])
        self.centre_coorsys_combobox.set_active(
                                            self.coorsys_index_map[coorsys])
        self.units_combobox.set_active(self.units_index_map['PIXEL'])

    def get_mode(self):
        return self.format_combobox.get_active()

    def set_mode(self, mode):
        self.format_combobox.set_active(mode)

    def _set_mode(self, mode):
        # @TODO: tooltips
        if mode == self.M_TOPLEFTBOTTOMRIGHT:
            self.notebook.set_current_page(0)
            self.pointA_label.set_text(name_to_label['topleft'])
            self.pointB_label.set_text(name_to_label['bottomright'])
        elif mode == self.M_TOPRIGHTBOTTOLEFT:
            self.notebook.set_current_page(0)
            self.pointA_label.set_text(name_to_label['topright'])
            self.pointB_label.set_text(name_to_label['bottomleft'])
        elif mode == self.M_CENTRESIZE:
            self.notebook.set_current_page(1)
        elif mode == self.M_VERTEX:
            self.notebook.set_current_page(2)
        else:
            raise ValueError('Invalid mode %s.' % mode)

    def get_corners_coorsys(self):
        return(self.corners_coorsys_combobox.get_active())

    def set_corners_coorsys(self, coorsys):
        self.corners_coorsys_combobox.set_active(
                                            self.coorsys_index_map[coorsys])

    def _set_corners_coorsys(self, coorsys):
        # @TODO: tooltips
        widgets = [getattr(self, 'point%s_spinbutton' % id_)
                                        for id_ in ('A1','A2','B1','B2')]
        if coorsys == 'ROWCOL':
            self.corners_coor1_label.set_text(
                                        self.coorsys_labels_map['ROWCOL'][0])
            self.corners_coor2_label.set_text(
                                        self.coorsys_labels_map['ROWCOL'][1])
            for widget in widgets:
                widget.set_digits(0)
                widget.set_range(0, 100000)
        elif coorsys == 'LATLON':
            self.corners_coor1_label.set_text(
                                        self.coorsys_labels_map['LATLON'][0])
            self.corners_coor2_label.set_text(
                                        self.coorsys_labels_map['LATLON'][1])
            for widget in widgets:
                widget.set_digits(4)
            # @TODO: check
            for widget in widgets[::2]:
                widget.set_range(-90.0, +90.0)
            for widget in widgets[1::2]:
                widget.set_range(-180.0, +180.0)
        else:
            raise ValueError('Invalid coordinate system %s, "ROWCOL" or '
                             '"LATLON" expected.' % mode)

    def get_centre_coorsys(self):
        return(self.centre_coorsys_combobox.get_active())

    def set_centre_coorsys(self, coorsys):
        self.centre_coorsys_combobox.set_active(self.coorsys_index_map[coorsys])

    def _set_centre_coorsys(self, coorsys):
        # @TODO: tooltips
        widgets = [getattr(self, 'centre_coor%s_spinbutton' % id_)
                                                        for id_ in (1, 2)]
        if coorsys == 'ROWCOL':
            self.centre_coor1_label.set_text(
                                    self.coorsys_labels_map['ROWCOL'][0])
            self.centre_coor2_label.set_text(
                                    self.coorsys_labels_map['ROWCOL'][1])
            for widget in widgets:
                widget.set_digits(0)
                widget.set_range(0, 100000)
        elif coorsys == 'LATLON':
            self.centre_coor1_label.set_text(
                                    self.coorsys_labels_map['LATLON'][0])
            self.centre_coor2_label.set_text(
                                    self.coorsys_labels_map['LATLON'][1])
            for widget in widgets:
                widget.set_digits(4)
                # @TODO: check
                self.centre_coor1_spinbutton.set_range(-90.0, +90.0)
                self.centre_coor1_spinbutton.set_range(-180.0, +180.0)
        else:
            raise ValueError('Invalid coordinate system %s, "ROWCOL" or '
                             '"LATLON" expected.' % mode)

    def get_vertex_coorsys(self):
        return self.vertexeditor.get_coorsys()

    def set_vertex_coorsys(self, coorsys):
        self.vertexeditor.set_coorsys(coorsys)

    def _set_vertex_coorsys(self, coorsys):
        self.vertexeditor.set_coorsys(coorsys)

    def get_units(self):
        return(self.units_index_map[self.units_combobox.get_active()])

    def set_units(self, units):
        self.units_combobox.set_active(self.units_index_map[units])

    def _set_units(self, units):
        # @TODO: tooltips
        widgets = [getattr(self,'%s_spinbutton' % id_)
                                        for id_ in ('width','height')]
        if units == 'PIXEL':
            for widget in widgets:
                widget.set_digits(0)
        elif units == 'KM':
            for widget in widgets:
                widget.set_digits(4)
        else:
            raise ValueError('Invalid units %s, "PIXEL" or "KM" expected.' % units)

    def get_values(self, vlist=None):
        funclist = ('_get_values_corners',
                    '_get_values_corners',
                    '_get_values_centresize',
                    '_get_values_vertex')
        indx = self.get_mode()
        func = getattr(self, funclist[indx])
        if vlist is None:
            vlist = []

        try:
            # Update the spin-button before getting the value
            toplevel = self.widget.get_toplevel()
            widget = toplevel.get_focus()
            widget.update()
        except AttributeError:
            pass

        return(func(vlist))

    def _get_values_corners(self, vlist=None):
        coorsys_combo = self.corners_coorsys_combobox
        widgets = [getattr(self, 'point%s_spinbutton' % id_)
                                            for id_ in ('A1','A2','B1','B2')]

        if coorsys_combo.get_active() == self.coorsys_index_map['ROWCOL']:
            values = [widget.get_value_as_int() for widget in widgets]
        else:
            values = [widget.get_value() for widget in widgets]

        coorsys_str = '"%s"' % coorsys_combo.get_active_text().upper()

        if vlist is None:
            vlist = []
        vlist.append( (name_to_label['coorsys'],  coorsys_str) )
        mode = self.get_mode()
        if mode == self.M_TOPLEFTBOTTOMRIGHT:
            vlist.append( (name_to_label['topleft'],
                                            ', '.join(map(str,values[:2]))) )
            vlist.append( (name_to_label['bottomright'],
                                            ', '.join(map(str,values[2:]))) )
        elif mode == self.M_TOPRIGHTBOTTOLEFT:
            vlist.append( (name_to_label['topright'],
                                            ', '.join(map(str,values[:2]))) )
            vlist.append( (name_to_label['bottomleft'],
                                            ', '.join(map(str,values[2:]))) )
        else:
            raise ValueError(('unexpected coordinate system mode: '
                              '"%s"') % self.mode_labels_map[mode])

        return vlist

    def _get_values_centresize(self, vlist=None):
        coorsys_combo = self.centre_coorsys_combobox
        widgets = [getattr(self, 'centre_coor%s_spinbutton' % id_)
                                                        for id_ in (1, 2)]

        if coorsys_combo.get_active() == self.coorsys_index_map['ROWCOL']:
            centrecoor = [widget.get_value_as_int() for widget in widgets]
        else:
            centrecoor = [widget.get_value() for widget in widgets]

        widgets = [getattr(self,'%s_spinbutton' % id_)
                                            for id_ in ('width','height')]
        if self.units_combobox.get_active() == self.units_index_map['PIXEL']:
            size = [widget.get_value_as_int() for widget in widgets]
        else:
            size = [widget.get_value() for widget in widgets]

        units_str = '"%s"' % self.units_combobox.get_active_text().upper()
        coorsys_str = '"%s"' % coorsys_combo.get_active_text().upper()

        if vlist is None:
            vlist = []
        vlist.append((name_to_label['coorsys'], coorsys_str))
        vlist.append((name_to_label['centre'], ', '.join(map(str, centrecoor))))
        vlist.append((name_to_label['sizeunit'], units_str))
        vlist.append((name_to_label['size'], ', '.join(map(str, size))))
        return vlist

    def _get_values_vertex(self, vlist=None):
        return self.vertexeditor.get_values(vlist)

    def set_values(self, vlist):
        '''Set values specified in vlist. vlist can be both a sequence of
        couples (key, value) and a dictionary like object.'''
        dict_ = dict(vlist)
        mode = self._get_mode_from_values(dict_)
        if mode is not None:
            self.set_mode(mode)
            funclist = ('_set_values_corners',
                        '_set_values_corners',
                        '_set_values_centresize',
                        '_set_values_vertex')
            fun = getattr(self, funclist[mode])
            return(fun(dict_))

    def _set_values_corners(self, values):
        # set the coordinate system
        key = name_to_label['coorsys'].lower()
        value = values.get(key)
        if value:
            value = remove_quote(value)
            try:
                self.corners_coorsys_combobox.set_active(
                                                self.coorsys_index_map[value])
            except KeyError, e:
                dialogs.run_message_dialog(
                    'Invalid coordinates system: %s' % e, gtk.MESSAGE_WARNING)

        # Set the corners
        for key, value in values.items():
            if key in (name_to_label['topleft'].lower(),
                       name_to_label['topright'].lower()):
                c1, c2 = value.split(',')
                self.pointA1_spinbutton.set_value(float(c1))
                self.pointA2_spinbutton.set_value(float(c2))
            elif key in (name_to_label['bottomright'].lower(),
                         name_to_label['bottomleft'].lower()):
                c1, c2 = value.split(',')
                self.pointB1_spinbutton.set_value(float(c1))
                self.pointB2_spinbutton.set_value(float(c2))

    def _set_values_centresize(self, values):
        # set the coordinate system
        key = name_to_label['coorsys'].lower()
        value = values.get(key)
        if value:
            value = remove_quote(value)
            try:
                self.centre_coorsys_combobox.set_active(
                                                self.coorsys_index_map[value])
            except KeyError, e:
                dialogs.run_message_dialog(
                    'Invalid coordinates system: %s' % e, gtk.MESSAGE_WARNING)

        # Set other params
        for key, value in values.items():
            if key == name_to_label['sizeunit'].lower():
                value = remove_quote(value)
                try:
                    self.centre_coorsys_combobox.set_active(
                                                self.units_index_map[value])
                except KeyError, e:
                    dialogs.run_message_dialog(
                        'Invalid coordinates system: %s' %
                                                e, gtk.MESSAGE_WARNING)
            elif key == name_to_label['centre'].lower():
                c1, c2 = value.split(',')
                self.centre_coor1_spinbutton.set_value(float(c1))
                self.centre_coor2_spinbutton.set_value(float(c2))
            elif key == name_to_label['size'].lower():
                c1, c2 = value.split(',')
                self.width_spinbutton.set_value(float(c1))
                self.height_spinbutton.set_value(float(c2))

    def _set_values_vertex(self, values):
        self.vertexeditor.set_values(values)

    def _get_mode_from_values(self, values):
        if values.has_key(name_to_label['topleft'].lower()):
            return(self.M_TOPLEFTBOTTOMRIGHT)
        if values.has_key(name_to_label['topright'].lower()):
            return(self.M_TOPRIGHTBOTTOLEFT)
        if values.has_key(name_to_label['centre'].lower()):
            return(self.M_CENTRESIZE)
        if values.has_key(name_to_label['vertexlist'].lower()):
            return(self.M_VERTEX)
        return None

    def clear(self):
        # Corners
        self.corners_coorsys_combobox.set_active(
                                        self.coorsys_index_map['ROWCOL'])
        #self.setCornersCoorSys('ROWCOL')
        self.pointA1_spinbutton.set_value(0)
        self.pointA2_spinbutton.set_value(0)
        self.pointB1_spinbutton.set_value(0)
        self.pointB2_spinbutton.set_value(0)

        # Centre and Size
        self.centre_coorsys_combobox.set_active(
                                        self.coorsys_index_map['ROWCOL'])
        #self.setCentreCoorSys('ROWCOL')
        self.centre_coorsys_combobox.set_active(self.units_index_map['PIXEL'])
        #self.set_units('PIXEL')
        self.centre_coor1_spinbutton.set_value(0)
        self.centre_coor2_spinbutton.set_value(0)
        self.width_spinbutton.set_value(0)
        self.height_spinbutton.set_value(0)

        # Vertex
        self.vertexeditor.set_coorsys('ROWCOL')
        self.vertexeditor.on_clear_all_items()
        self.vertexeditor.backup.clear()

    def set_default(self):
        self.clear()

        # Corners
        self.pointB1_spinbutton.set_value(100)
        self.pointB2_spinbutton.set_value(100)

        # Centre and Size
        self.centre_coor1_spinbutton.set_value(100)
        self.centre_coor2_spinbutton.set_value(100)
        self.width_spinbutton.set_value(100)
        self.height_spinbutton.set_value(100)

        # Vertex
        self.vertexeditor.set_values(self.vertexeditor.defaultvalues)

        # Mode
        self.format_combobox.set_active(self.M_TOPLEFTBOTTOMRIGHT)

    def set_from_cfg(self, cfg, section, errors_='strict'):
        labels = [label.lower() for label in name_to_label.values()]
        values = {}
        for label in labels:
            if cfg.has_option(section, label):
                try:
                    values[label] = cfg.get(section, label)
                except ValueError:
                    if errors_ == 'strict':
                        raise
                    elif errors_ == 'warn':
                        msg = ('Invalid value "%s" for the "%s" '
                               'parameter.\nIgnoreg.') % (val, par.name)
                        dialog = gtk.MessageDialog(
                            #self.widget.get_toplevel(),
                            flags=gtk.DIALOG_MODAL |
                                  gtk.DIALOG_DESTROY_WITH_PARENT,
                            type=gtk.MESSAGE_WARNING,
                            buttons=gtk.BUTTONS_CLOSE,
                            message_format=msg)
                        dialog.run()
                        dialog.destroy()
        if values:
            self.set_values(values)

    def on_AoI_format_changed(self, widget, data=None):
        self._set_mode(self.format_combobox.get_active())

    def on_corners_coorsys_changed(self, widget, data=None):
        indx = self.corners_coorsys_combobox.get_active()
        self._set_corners_coorsys(self.coorsys_index_map[indx])

    def on_centre_coorsys_changed(self, widget, data=None):
        indx = self.centre_coorsys_combobox.get_active()
        self._set_centre_coorsys(self.coorsys_index_map[indx])

    def on_units_changed(self, widget, data=None):
        self._set_units(self.units_index_map[self.units_combobox.get_active()])

class AoIDialog(gladewrap.GladeDialog, AoIForm):
    '''Dialog for AoI selection.'''

    root = 'AoI_dialog'

    def __init__(self, parent=None):
        AoIForm.__init__(self)
        # @TODO: __init__ method from base class 'gladewrap.GladeDialog'
        #        is not called
        if parent:
            self.set_parent(parent)
        self.set_default_response(gtk.RESPONSE_OK)
        self.set_size_request(450, -1)

if(__name__ == '__main__'):
    def test_items_editor_dialog():
        dialog = gtk.Dialog('Items Editor', None,
                            buttons = (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        itemseditor = ItemsEditor()
        itemseditor.set_values(['a', 'b', 'c', 'd', 'e'])
        dialog.vbox.pack_start(itemseditor.widget)
        dialog.set_size_request(450, 300)
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            print '\n'.join([str(item) for item in itemseditor.get_values()])
        dialog.destroy()

    def test_vertex_editor_dialog():
        #~ d = gtk.Dialog('Vertex Editor', None,
                        #~ buttons = (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                   #~ gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        #~ vertexEditor = VertexEditor()
        #~ d.vbox.pack_start(vertexEditor.widget)
        #~ d.set_size_request(450, 300)
        dialog = VertexEditorDialog()
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            print '\n'.join(['%s = %s' % (key, val)
                                    for key, val in dialog.get_values()])
        dialog.destroy()

    def test_AoI_dialog():
        import StringIO

        inifile_data = '''
[AOI]
Coordinate System = "ROWCOL"
Top Left Corner = 1, 2
Bottom Right Corner = 3, 4
'''
        inifile = StringIO.StringIO(inifile_data)

        dialog = AoIDialog()

        if inifile:
            import ConfigParser
            cfg = ConfigParser.SafeConfigParser()
            cfg.readfp(inifile)
            dialog.set_from_cfg(cfg, 'AOI')

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            print '\n'.join(['%s = %s' % (key, val)
                                        for key, val in dialog.get_values()])
        dialog.destroy()

    os.chdir(os.path.expanduser('~/tmp'))
    #test_ItemsEditorDialog()
    #~ test_VertexEditorDialog()
    test_AoI_dialog()
