### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''Tools for running external processes.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__version__ = '$Revision: 704 $'
__date__    = '$Date: 2009-04-13 10:09:18 +0200 (lun, 13 apr 2009) $'

import re
import os
import sys
import time
import logging
import subprocess2
from cStringIO import StringIO

if sys.platform[:3] == 'win':
    EX_OK = 0
else:
    EX_OK = os.EX_OK

# @TODO: new BaseOutputHandler class and rename the current BaseOutputHandler
#        TextOutputHandler
class BaseOutputHandler(object):

    def __init__(self):
        super(BaseOutputHandler, self).__init__()
        self._buffer = StringIO()
        self._wpos  = self._buffer.tell()

        self.error_flag = False

        # NOTE: progress, percentage and text are all optional. The regexp
        #       consumes the '\r' character in any case.
        pattern = ('[ \t]'
                   '*\r+'
                   '[ \t]*'
                   '(?P<pulse>[\\\|/-])?'
                   '[ \t]*'
                   '((?P<percentage>\d{1,3}(\.\d*)?)[ \t]*%)?'
                   '[ \t]*'
                   '((?P<text>.*?)(?=\r|\n))')
        self._progress_pattern = re.compile(pattern)

        self.handlers = ['progress', 'line']

        self.percentage_fmt = '%5.1f %%'

    def reset(self):
        self._buffer.close()
        self._buffer = StringIO()
        self._wpos = self._buffer.tell()
        self.error_flag = False

    def close(self):
        self._parse()
        data = self._buffer.read()
        if data:
            if not data.endswith('\n'):
                data += '\n'
            self.handle_line(data)
        self.reset()

    def feed(self, data):
        rpos = self._buffer.tell()
        self._buffer.seek(self._wpos)
        self._buffer.write(data)
        self._wpos = self._buffer.tell()
        self._buffer.seek(rpos)
        self._parse()

    def get_progress(self):
        pos = self._buffer.tell()
        data = self._buffer.read()
        match = self._progress_pattern.match(data)
        if match:
            result = [match.group('pulse'), match.group('percentage'), match.group('text')]
            if result == [None, None, None]:
                result = None
        else:
            result = None

        if result:
            self._buffer.seek(pos+match.end())
        else:
            self._buffer.seek(pos)
            return None

        if result[1] is not None:
            result[1] = float(result[1])
        return result

    def get_line(self):
        pos = self._buffer.tell()
        data = self._buffer.readline()
        if data and (data[-1] == '\n'):
            return data
        self._buffer.seek(pos)
        return None

    def _parse(self):
        while True:
            for name in self.handlers:
                data = getattr(self, 'get_' + name)()
                if data not in (None, ''):
                    getattr(self, 'handle_' + name)(data)
                    break
            else:
                # no pattern matches
                break

    def handle_progress(self, data):
        pulse, percentage, text = data
        result = []
        if pulse:
            result.append(pulse)
        if percentage is not None:
            result.append(self.percentage_fmt % percentage)
        if text:
            result.append(text)
        sys.stdout.write('\r' + ' '.join(result))
        sys.stdout.flush()

    def handle_line(self, data):
        sys.stdout.write(data)
        sys.stdout.flush()

class BaseExecutor(object):
    '''Base class for command line tools executors.'''

    def __init__(self, outputhandler=None, logger=None):
        super(BaseExecutor, self).__init__()
        self.subprocess = None
        self.shell = False
        self._stopped = False
        self._error_flag = False

        if not outputhandler:
            self.outputhandler = BaseOutputHandler()
        else:
            self.outputhandler = outputhandler

        if not logger:
            self.logger = logging.getLogger()
        else:
            self.logger = logger

    def get_error_flag(self):
        if self.outputhandler.error_flag:
            self._error_flag = True
            return True
        if self.subprocess:
            return (self._error_flag or
                    self.subprocess.returncode not in (None, EX_OK))
        return self._error_flag

    error_flag = property(get_error_flag)

    def finalize_hook(self):
        pass

    def finalize(self):
        try:
            if self.subprocess:
                # retrieve residual data
                if sys.platform[:3] == 'win':
                    # using read() here hangs on win32
                    data = self.subprocess.recv()
                    while data:
                        self.outputhandler.feed(data)
                        data = self.subprocess.recv()
                else:
                    try:
                        data = self.subprocess.stdout.read()
                        self.outputhandler.feed(data)
                    except ValueError:
                        # I/O operation on closed file.
                        pass

                # close the pipe and wait for the subprocess termination
                if self.subprocess.stdout:
                    # subprocess2 closes the stdout if no more output is
                    # available
                    self.subprocess.stdout.close()
                self.subprocess.wait()
                if self.outputhandler.error_flag:
                    # Save the error_flag status
                    self._error_flag = True
                self.outputhandler.close()

                if self.subprocess.returncode != EX_OK:
                    if self._stopped:
                        self.logger.info('Execution stopped by the user.')
                    else:
                        msg = ('Process (PID=%d) exited with return code %d.' %
                                               (self.subprocess.pid,
                                                self.subprocess.returncode))
                        self.logger.warning(msg)

                # Call finalize hook is available
                self.finalize_hook()
        finally:
            # Protect for unexpected errors in the feed and close methods of
            # the outputhandler
            self.reset()

    def reset(self):
        if self.subprocess:
            self.subprocess.stop(force=True)

        assert (self.subprocess is None or
                self.subprocess.returncode is not None), \
                                        'the process is still running'
        self.outputhandler.reset()
        self.subprocess = None
        self._stopped = False
        self._error_flag = False

    def prerun_hook(self, cmd):
        if sys.platform[:3] == 'win':
            prompt = 'cmd >'
        else:
            prompt = '$'

        if not isinstance(cmd, basestring):
            cmd = ' '.join(cmd)

        self.logger.info('%s %s\n' % (prompt, cmd))

    def connect_output_handler(self):
        while self.subprocess:
            self.handle_output()
            time.sleep(0.1)

    def handle_output(self, *args, **kwargs):
        if self.subprocess is None or self.subprocess.poll() is not None:
            # NOTE: 'self.subprocess is None' shold never happen ar this point
            self.finalize()
            return False
        else:
            data = self.subprocess.recv()
            if data:
                self.outputhandler.feed(data)
            return True

    def run(self, args, cwd=None, env=None):
        assert self.subprocess is None
        if sys.platform[:3] == 'win':
            closefds = False
            startupinfo = subprocess2.STARTUPINFO()
            startupinfo.dwFlags |= subprocess2.STARTF_USESHOWWINDOW
        else:
            closefds = True
            startupinfo = None

        self.outputhandler.reset()
        self.prerun_hook(args)

        try:
            self.subprocess = subprocess2.Popen(args,
                                                stdin = subprocess2.PIPE,
                                                stdout = subprocess2.PIPE,
                                                stderr = subprocess2.STDOUT,
                                                cwd = cwd,
                                                env = env,
                                                close_fds = closefds,
                                                shell = self.shell,
                                                startupinfo = startupinfo)

            self.subprocess.stdin.close()
            self.connect_output_handler()
        except OSError, e:
            if not isinstance(args, basestring):
                args = ' '.join(args)
            msg = 'Unable to execute: "%s"' % args
            self.logger.error(msg, exc_info=True)
            self.reset()
        except:
            self.reset()
            raise

    def stop(self):
        if self.subprocess:
            if self.subprocess.poll() is not None:
                # If an unexpected error happens in the handle_output method
                # no more watch events are scheduled and the finalize mathod
                # is never called. This should fix the issue.
                self.finalize()

            self.logger.debug('Execution stopped by the user.')
            self._stopped = True
            stopped = self.subprocess.stop()
            if not stopped:
                msg = ('Unable to stop the sub-process (PID=%d).' %
                                                        self.subprocess.pid)
                self.logger.warning(msg)
                self.reset()
            # The subprocess is successfully stopped.
            # The output handler will provide to the finalization
        else:
            self.reset()

if(__name__ == '__main__'):
    h = BaseOutputHandler()
    h.feed('\rline\n')
    h.feed('\r-')
    h.feed(' \r/')
    h.feed('\r \\')
    h.feed('\r| ')
    h.feed(' \r - ')

    h.feed('\rline\n')
    h.feed('\r12%')
    h.feed(' \r13%')
    h.feed('\r 14%')
    h.feed('\r15 %')
    h.feed('\r16% ')
    h.feed(' \r 17 % ')

    h.feed('\rline\n')
    h.feed('\r17.1%')
    h.feed(' \r17.2%')
    h.feed('\r 17.3%')
    h.feed('\r17.4 %')
    h.feed('\r17.5% ')
    h.feed(' \r 17.6 % ')

    h.feed('\rline\n')
    h.feed('\r-18%')
    h.feed(' \r/18%')
    h.feed('\r |18% ')
    h.feed('\r\\ 18%')
    h.feed('\r-18 %')
    h.feed('\r/18% ')
    h.feed(' \r   |   18 % ')

    h.feed('\rline\n')
    h.feed('\r-19.0%')
    h.feed(' \r/19.1%')
    h.feed('\r |19.2%')
    h.feed('\r\\ 19.3%')
    h.feed('\r-19.4 %')
    h.feed('\r/19.5% ')
    h.feed(' \r   |   19.5 % ')

    h.feed('\rline\n')
    h.feed('\relapsed time: 1')
    h.feed('\relapsed time: 2')
    h.feed('\relapsed time: 3')

    h.feed('\rline\n')
    h.feed('\r- 1.0% completed.')
    h.feed('\r/ 2.0% completed.')
    h.feed('\r| 3.0% completed.')
    h.feed('\r\\ 4.0% completed.')

    h.feed('\rdone\n')
    h.close()

    #~ import doctest
    #~ doctest.testmod()
