### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''Tools for running external processes in a GTK GUI.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__version__ = '$Revision: 704 $'
__date__    = '$Date: 2009-04-13 10:09:18 +0200 (lun, 13 apr 2009) $'

import re
import os
import time
import locale
import logging

import gtk
import gobject

from exectools import BaseOutputHandler, BaseExecutor

class GtkBlinker(gtk.Image):
    def __init__(self):
        gtk.Image.__init__(self)
        self.set_from_stock(gtk.STOCK_MEDIA_RECORD, gtk.ICON_SIZE_SMALL_TOOLBAR)

    def pulse(self):
        sensitive = self.get_property('sensitive')
        sensitive = not sensitive
        self.set_sensitive(sensitive)

    def flush(self):
        while gtk.events_pending():
            gtk.main_iteration(False)

class GtkStream(object):
    def __init__(self, textview=None):
        super(GtkStream, self).__init__()
        if textview is None:
            textview = gtk.TextView()
        self.textview = textview
        self.srcencoding = locale.getpreferredencoding()

    def fixencoding(self, data):
        # @TODO: test on win32
        if self.srcencoding != 'UTF-8':
            try:
                data = data.decode(self.srcencoding, 'replace')
                data = data.encode('UTF-8', 'replace')
            except UnicodeDecodeError, e:
                data = str(e)
                logging.warning('unicode error.', exc_info=True)
        return data

    def fixtags(self, tags):
        buf = self.textview.get_buffer()
        tagtable = buf.get_tag_table()
        tags = [tag for tag in tags if tagtable.lookup(tag)]
        return tuple(tags)

    def flush(self):
        while gtk.events_pending():
            gtk.main_iteration(False)

    def write(self, data, *tags):
        data = self.fixencoding(data)
        tags = self.fixtags(tags)

        buf = self.textview.get_buffer()
        textiter = buf.get_end_iter()
        buf.insert_with_tags_by_name(textiter, data, *tags)

        buf.place_cursor(buf.get_end_iter())
        self.textview.scroll_mark_onscreen(buf.get_mark('insert'))

class GtkOutputPlane(GtkStream):

    def __init__(self, textview, hide_callback=None):
        super(GtkOutputPlane, self).__init__(textview)
        textview.connect('populate-popup', self.on_populate_popup)

        self.hide_callback = hide_callback

        self.filedialog = self.setup_filedialog()

    def setup_filedialog(self):
        dialog = gtk.FileChooserDialog(
                        title   = 'Save Output Log',
                        parent  = self.textview.get_toplevel(),
                        action  = gtk.FILE_CHOOSER_ACTION_SAVE,
                        buttons = (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                   gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        patterns = [('*.txt', 'Text files'), ('*', 'All Files')]
        for pattern, name in patterns:
            filefilter = gtk.FileFilter()
            filefilter.set_name(name)
            filefilter.add_pattern(pattern)
            dialog.add_filter(filefilter)

        dialog.set_current_name('outputlog.txt')
        dialog.set_select_multiple(False)
        dialog.set_default_response(gtk.RESPONSE_OK)

        # @COMPATIBILITY: gtk.FileChooser.set_do_overwrite_confirmation is
        #                 avalilable in PyGTK 2.8 and above
        if gtk.pygtk_version >= (2,8,0):
            dialog.set_do_overwrite_confirmation(True)

        return dialog

    def report(self):
        header = '# Output log generated on %s\n\n' % time.asctime()

        buf = self.textview.get_buffer()
        text = buf.get_text(buf.get_start_iter(), buf.get_end_iter())

        return header + text

    def clear(self):
        buf = self.textview.get_buffer()
        buf.set_text('')

    def save(self):
        dialog = self.filedialog
        dialog.set_transient_for(self.textview.get_toplevel())

        # @COMPATIBILITY: gtk.FileChooser.set_do_overwrite_confirmation is
        #                 avalilable in PyGTK 2.8 and above
        if gtk.pygtk_version >= (2,8,0):
            response = dialog.run()
            if response == gtk.RESPONSE_CANCEL:
                dialog.hide()
                return
            filename = dialog.get_filename()
        else:
            filename = None
            while not filename:
                response = dialog.run()
                if response == gtk.RESPONSE_CANCEL:
                    dialog.hide()
                    return
                filename = dialog.get_filename()
                if filename and os.path.exists(filename):
                    msg = ('File "%s" already exists.\n\n'
                           'Are you sure you want overwrite it?' % filename)
                    msgdialog = gtk.MessageDialog(
                                        parent  = dialog,
                                        flags   = gtk.DIALOG_MODAL |
                                                  gtk.DIALOG_DESTROY_WITH_PARENT,
                                        type    = gtk.MESSAGE_QUESTION ,
                                        buttons = gtk.BUTTONS_YES_NO,
                                        message_format = msg)
                    msgdialog.set_default_response(gtk.RESPONSE_NO)
                    response = msgdialog.run()
                    msgdialog.destroy()
                    if(response != gtk.RESPONSE_YES):
                        filename = None

        dialog.hide()

        logfile = open(filename, 'w')
        text = self.report()
        logfile.write(text)
        logfile.close()

    def on_populate_popup(self, widget, menu, data=None):
        # Separator
        item = gtk.SeparatorMenuItem()
        item.set_name('separator')
        item.show()
        menu.append(item)

        # Save As
        item = gtk.ImageMenuItem(gtk.STOCK_SAVE_AS)
        item.set_name('save_as')
        item.connect('activate', self.on_save, None)
        item.show()
        menu.append(item)

        # Clear OutputLog
        item = gtk.ImageMenuItem(gtk.STOCK_CLEAR)
        item.set_name('clear')
        item.connect('activate', self.on_clear, None)
        item.show()
        menu.append(item)

        # Hide OutputLog
        if self.hide_callback:
            item = gtk.ImageMenuItem(gtk.STOCK_CLOSE)
            item.set_name('close')
            item.connect('activate', self.hide_callback)
            item.show()
            menu.append(item)

    def on_clear(self, widget=None, data=None):
        self.clear()

    def on_save(self, widget=None, data=None):
        self.save()

class GtkOutputHandler(BaseOutputHandler, GtkStream):
    def __init__(self, textview, statusbar, progressbar=None, blinker=None):
        BaseOutputHandler.__init__(self)
        GtkStream.__init__(self, textview)

        self.context_id = statusbar.get_context_id('progress')
        self.statusbar = statusbar

        if progressbar is None:
            progressbar = gtk.ProgressBar()
            statusbar.pack_end(progressbar)
            progressbar.hide()
        self.progressbar = progressbar

        if blinker is None:
            blinker = GtkBlinker()
            statusbar.pack_end(blinker, expand=False)
            blinker.hide()
        self.blinker = blinker

        self.texttags = {
            'error': re.compile('error', re.IGNORECASE),
            'warning': re.compile('warning', re.IGNORECASE),
        }

    def feed(self, data):
        self.blinker.show()
        #self.progressbar.show()
        super(GtkOutputHandler, self).feed(data)

    def close(self):
        self.statusbar.pop(self.context_id)
        super(GtkOutputHandler, self).close()

    def reset(self):
        super(GtkOutputHandler, self).reset()
        self.progressbar.set_text('%.0f %%' % 0)
        self.progressbar.hide()
        self.progressbar.set_fraction(0)
        self.blinker.hide()

    def _handle_pulse(self, data):
        self.blinker.pulse()

    def _handle_percentage(self, data):
        self.progressbar.show()
        self.progressbar.set_text(self.percentage_fmt % data)
        self.progressbar.set_fraction(data / 100.)

    def handle_progress(self, data):
        pulse, percentage, text = data
        if pulse:
            self._handle_pulse(pulse)
        if percentage is not None:
            self._handle_percentage(percentage)
            #self.flush()
        if text and not pulse and percentage is None:
            self.statusbar.pop(self.context_id)
            self.statusbar.push(self.context_id, text)
            self._handle_pulse(pulse)
            #self.flush()
        self.flush() # might slow too mutch

    def handle_line(self, data):
        for tag_name, pattern in self.texttags.items():
            match = pattern.search(data)
            if match:
                if self.match_hook(data, tag_name, match):
                    self.write(data, tag_name)
                break
        else:
            self.write(data)
        self.flush()

    def match_hook(self, data, tag_name, match_obj):
        if tag_name == 'error':
            self.error_flag = True
        return True

class GtkStreamLoggingHandler(logging.StreamHandler):
    level2tag = {
        logging.CRITICAL: 'error',
        # FATAL = CRITICAL
        logging.ERROR: 'error',
        logging.WARNING: 'warning',
        # WARN = WARNING
        logging.INFO: 'info',
        logging.DEBUG: 'debug',
        logging.NOTSET: '',
    }

    def __init__(self, strm=None):
        if strm is None:
            strm = GtkStream()

        if isinstance(strm, gtk.TextView):
            strm = GtkStream(strm)

        assert isinstance(strm, GtkStream)
        logging.StreamHandler.__init__(self, strm)

    def emit(self, record):
        try:
            msg = self.format(record)
            tag = self.level2tag.get(record.levelno, '')
            self.stream.write('%s\n' % msg, tag)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

class GtkDialogLoggingHandler(logging.Handler):
    levelsmap = {
        logging.CRITICAL: gtk.MESSAGE_ERROR,
        # FATAL = CRITICAL
        logging.ERROR: gtk.MESSAGE_ERROR,
        logging.WARNING: gtk.MESSAGE_WARNING,
        # WARN = WARNING
        logging.INFO: gtk.MESSAGE_INFO,
        logging.DEBUG: gtk.MESSAGE_INFO,
        logging.NOTSET: gtk.MESSAGE_INFO,
    }

    def __init__(self, dialog=None, parent=None):
        logging.Handler.__init__(self)
        if dialog is None:
            if parent is None:
                try:
                    parent = gtk.window_list_toplevels()[0]
                except IndexError:
                    pass
            dialog = gtk.MessageDialog(parent, buttons=gtk.BUTTONS_CLOSE)
        self.dialog = dialog
        self.formatter = None

    def emit(self, record):
        try:
            msgtype = self.levelsmap[record.levelno]
            self.dialog.set_property('message-type', msgtype)

            msg = self.format(record)
            msg = msg.encode('UTF-8', 'replace')

            # NOTE: format_secondary_markup requires PyGTK 2.6
            self.dialog.format_secondary_markup(msg)
            if record.exc_info:
                msg = record.getMessage()
                msg = msg.encode('UTF-8', 'replace')
                self.dialog.set_markup('<b>%s</b>' % msg)
            else:
                msg = logging.getLevelName(record.levelno)
                self.dialog.set_markup('<b>%s</b>' % msg)

            self.dialog.run()
            self.dialog.hide()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

class GtkTimeoutExecutor(BaseExecutor):

    # @TODO: make this configurable
    timeout = 50 # ms

    def connect_output_handler(self):
        gobject.timeout_add(self.timeout, self.handle_output)

class GtkIOWatchExecutor(BaseExecutor):
    def connect_output_handler(self):
        # @TODO: use separate handlers for IO_IN | IO_PRI and IO_ERR | IO_HUP

        cond = gobject.IO_IN | gobject.IO_PRI | gobject.IO_ERR | gobject.IO_HUP
        gobject.io_add_watch(self.subprocess.stdout, cond, self.handle_output)

    def handle_output(self, fd, condition):
        #assert fd is self.subprocess.stdout
        #assert self.subprocess is not None
        if(condition & gobject.IO_IN) or (condition & gobject.IO_PRI):
            # Use the non blocking read from subprocess2
            data = self.subprocess.recv()

            ## As an alternative one can use the satndard file object read.
            ## Read just one cherecter at the time in order to avoid blocking
            ## if not enough input is available for the I/O.
            ## The brawback is that the handle_output method is called a
            ## number of times equal to the number of characters that the
            ## subprocess outputs
            #data = fd.read(1)

            self.outputhandler.feed(data)

        if(condition & gobject.IO_ERR) or (condition & gobject.IO_HUP):
            # finalize also closes the fd
            self.finalize()
            return False
        return True
