### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''BEST file patterns.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__date__    = '$Date: 2009-04-13 10:09:18 +0200 (lun, 13 apr 2009) $'
__version__ = '$Revision: 704 $'

import os
import fnmatch

class FilePattern(object):
    def __init__(self, label, patterns):
        super(FilePattern, self).__init__()
        self.label = label
        self.value = patterns

_patternsmap = {}
_exclude = ['.HAN', '.BG', '.SDf', '.txt', '.tif', '.ers']
_all_patterns = set()

def set_patterns_map(pmap):
    _patternsmap.clear()
    _all_patterns.clear()

    _patternsmap.update(pmap)
    need_fix = []
    for tool in _patternsmap:
        pattern = _patternsmap[tool].value
        if isinstance(pattern, tuple):
            _all_patterns.update(pattern)
        elif pattern == 'ALL':
            need_fix.append(tool)
    _all_patterns.difference_update(_exclude)

    for tool in need_fix:
        _patternsmap[tool].value = tuple(_all_patterns)

def fix_patterns(patterns):
    '''Expand patterns in glob form

    >>> set_patterns_map(scan_config())
    >>> fix_patterns('*.??r')
    '*.XTr *.ITr *.GTr'

    >>> fix_patterns('*.APf')
    '*.APf'
    '''
    result = []
    for pattern in patterns.split():
        matching = fnmatch.filter(_all_patterns, pattern)
        result.extend(['*' + ext for ext in matching])
    if result:
        return ' '.join(result)
    else:
        return patterns

def find_patterns(filters):
    '''Find file patterns'''

    bestpatterns = _patternsmap.values()
    extra_patterns = dict()
    for filter_ in filters:
        patterns = filter_[0]
        for pattern in patterns.split():
            if pattern.lstrip('*') in _all_patterns:
                continue
            for filepattern in bestpatterns:
                matching = []
                for value in filepattern.value:
                    if value in _all_patterns and fnmatch.fnmatch(value, pattern):
                        matching.append(value)
                if matching:
                    value = extra_patterns.setdefault(filepattern.label, [])
                    value.extend(matching)

    result = [(fix_patterns(patterns), description)
                            for patterns, description in filters]

    for description, patterns in extra_patterns.items():
        patterns.sort()
        if len(patterns) < 5:
            description = '%s (%s)' % (description, ' '.join(patterns))
        patterns = ' '.join(['*' + pattern for pattern in patterns])
        result.append((patterns, description))

    return result

def get_ext(tool, infilename=None):
    exts = _patternsmap[tool].value
    extslen = len(exts)
    if extslen == 1:
        return exts[0]
    elif extslen > 1:
        if infilename:
            inputfile_ext = os.path.splitext(infilename)[-1]
            lastchar = inputfile_ext[-1]
            outext = [e for e in exts if e[-1] == lastchar]
            if outext:
                return outext[0]
            else:
                # outext is false now
                if lastchar in 'isr':
                    outext = [e for e in exts if e[-1] == 'f']
                elif lastchar == 't':
                    outext = [e for e in exts if e[-1] == 'c']
                if outext:
                    return outext[0]
    return None

def scan_config():
    import os
    import glob
    import utils
    import ConfigParser
    configdir = utils._configdir # @TODO: fix
    patternsmap = {}
    for hmifilename in glob.glob(os.path.join(configdir, 'hmi', '*.ini')):
        try:
            cfg = ConfigParser.SafeConfigParser()
            cfg.read(hmifilename)

            title    = cfg.get('MAIN', 'title')
            label    = cfg.get('MAIN', 'patterns_label')
            patterns = cfg.get('MAIN', 'patterns')
            if patterns != 'ALL':
                patterns = tuple([item.strip()
                                        for item in patterns.split(',')])

            patternsmap[title] = FilePattern(label, patterns)

        except ConfigParser.NoSectionError, e:
            e.args.append('hmifilename=%s' % hmifilename)
            raise

    return patternsmap

#~ ALL_IMAGES     = (ALL, 'BEST Internal Format (.??[cifrst])')
#~ L1_IMAGES      = (fix_patterns('*.??[cifst]'), 'Level 1 Images (.??[cifst])')
#~ REAL_IMAGES    = (fix_patterns('*.??[fis]'),   'Real Images (.??f, .??i, .??s)')
#~ COMPLEX_IMAGES = (fix_patterns('*.??[crt]'),   'Complex Images (.??c, .??r, .??t)')
#~ L1_COMPLEX_IMAGES = (fix_patterns('*.??[ct]'), 'Level 1 Complex Images (.??c, .??t)')
#~ INT_IMAGES     = (fix_patterns('*.??[irst]'),  'Integer Images (.??[irst])')
#~ IMT8_IMAGES    = (fix_patterns('*.??[ir]'),    'Real/Complex 8-bit Integer Images (.??i, .??r)')
#~ INT16_IMAGES   = (fix_patterns('*.??[st]'),    'Real/Complex 16-bit Integer Images (.??s, .??t)')
#~ FLOAT_IMAGES   = (fix_patterns('*.??[cf]'),    'Real/Complex Float Images (.??c, .??f)')
#~ R_IMAGES = (fix_patterns('*.??r'), 'RAW Images, 8 bits + 8 bits  (.??r)')
#~ I_IMAGES = (fix_patterns('*.??i'), '8-bit Integer Images (.??i)')
#~ S_IMAGES = (fix_patterns('*.??s'), '16-bit Integer Images (.??s)')
#~ T_IMAGES = (fix_patterns('*.??t'), 'Complex Integer Images, 16 bits + 16 bits (.??t)')
#~ F_IMAGES = (fix_patterns('*.??f'), '32-bit Float Images (.??f)')
#~ C_IMAGES = (fix_patterns('*.??c'), 'Complex Float Images, 32 bits + 32 bits (.??c)')

if __name__ == '__main__':
    def _test():
        import doctest
        doctest.testmod(verbose=True)

    _test()
