### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''Forms for BEST parameters files setup and management.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__version__ = '$Revision: 704 $'
__date__    = '$Date: 2009-04-13 10:09:18 +0200 (lun, 13 apr 2009) $'

import os
import warnings
import ConfigParser

import gtk

import utils
import errors
import dialogs
import gladewrap
import paramtype
import aoi

# -----------------------------------------------------------------------------
# --- Param set containers ----------------------------------------------------
# -----------------------------------------------------------------------------
class BaseParamSetDecoration(object):
    '''Base class for ParamSet decoration.

    This class and its derivates are designed to provide decorations (title,
    scroll-bars and so on) to ParamSet classes.
    It is thought to be used as MixIn class together with a ParamSet class and
    it expects that some attribures/methods like self.label, self.clear and
    self.set_default are defined.
    The user shoult inherit from both a ParamSet and one ParamSetContainer
    and use the addContainer method in the inizializer.
    '''

    def _decorate(self, widget, buttons=False):
        '''Build the decoration.

        Build the decoration and pack the "widget" parameter inside it.
        If the "buttons" parameter evaluates True a button-bar with the "Clean"
        and "Dafault" buttons is added to the decoration.
        The return value is the most extern widget.
        '''

        raise NotImplementedError('%s is a virtual class.' %
                                                    self.__class__.__name__)

    def on_clear(self, widget=None):
        '''Callback for the ParamSet "clear" method.'''
        self.clear()

    def on_set_default(self, widget=None):
        '''Callback for  the ParamSet "set_default" method.'''
        self.set_default()

    def buttonbox(self):
        '''Build a button-bar with the "Clean" and the "Default" buttons.'''
        gladefile = utils.get_gladefile('bestgui.glade')
        gui = gladewrap.GladeObject(gladefile, 'paramset_buttonbox')
        dict_ = {
            'on_clear':      self.on_clear,
            'on_set_default': self.on_set_default,
        }
        gui.xml.signal_autoconnect(dict_)
        return gui.widget

    # @TODO: improve the api
    #~ def getTitle(self): return self._frame.get_label()
    #~ def setTitle(self, title):
        #~ self._frame.set_label(self.title_format % self.label)
    #~ title = property(getTitle, setTitle, None, 'Set the container title')

class MainSet(BaseParamSetDecoration):
    '''Decoration for main parameters sets.

    This class is designed to provide decorations to ParamSet classes.
    The decoration is a frame with a centered title in bold large characters
    and scrollbars.
    MainSet is thought to be used as MixIn class together with a ParamSet
    class and it expects that some attribures/methods like self.label,
    self.clear and self.set_default are defined.
    The user shoult inherit from both a ParamSet and one ParamSetContainer
    and use the addContainer method in the inizializer.
    '''

    title_format = '<span size="xx-large"><b>%s</b></span>'

    def _decorate(self, widget, buttons=False):
        '''Build the decoration.

        Build the decoration and pack the "widget" parameter inside it.
        If the "buttons" parameter evaluates True a button-bar with the "Clean"
        and "Dafault" buttons is added to the decoration.
        The return value is the most extern widget.
        '''

        self._frame = gtk.Frame(self.title_format % self.label)
        self._frame.set_label_align(0.5, 0.5)
        self._frame.get_label_widget().set_use_markup(True)
        self._frame.set_shadow_type(gtk.SHADOW_NONE)
        scrolledwin = gtk.ScrolledWindow()
        scrolledwin.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolledwin.add_with_viewport(widget)
        if buttons:
            bbox = self.buttonbox()
            vbox = gtk.VBox()
            vbox.pack_start(scrolledwin)
            vbox.pack_start(bbox, expand=False)
            self._frame.add(vbox)
        else:
            self._frame.add(scrolledwin)
        return self._frame

class BaseSubSet(BaseParamSetDecoration):
    '''Decoration for base sub-set of parameters.

    This class is designed to provide decorations to ParamSet classes.
    The decoration is a frame with a bold title.
    BaseSubSet is thought to be used as MixIn class together with a ParamSet
    class and it expects that some attribures/methods like self.label,
    self.clear and self.set_default are defined.
    The user shoult inherit from both a ParamSet and one ParamSetContainer
    and use the addContainer method in the inizializer.
    '''

    title_format = '<b>%s</b>'

    def _decorate(self, widget, buttons=True):
        '''Build the decoration.

        Build the decoration and pack the "widget" parameter inside it.
        If the "buttons" parameter evaluates True a button-bar with the "Clean"
        and "Dafault" buttons is added to the decoration.
        The return value is the most extern widget.
        '''

        self._frame = gtk.Frame(self.title_format % self.label)
        label_widget = self._frame.get_label_widget()
        label_widget.set_use_markup(True)
        label_widget.set_padding(6, 2)
        #self._frame.set_shadow_type(gtk.SHADOW_NONE)
        if buttons:
            bbox = self.buttonbox()
            vbox = gtk.VBox()
            #vbox.pack_start(widget)
            vbox.pack_start(bbox, expand=False)
            vbox.pack_start(widget)
            self._frame.add(vbox)
        else:
            self._frame.add(self.widget)
        return self._frame

class SubSet(BaseSubSet):
    '''Decoration for sub-set of parameters.

    This class is designed to provide decorations to ParamSet classes.
    The decoration is a frame with a bold title and a checkbutton that allows
    to disble the sub-set.
    SubSet is thought to be used as MixIn class together with a ParamSet class
    and it expects that some attribures/methods like self.label, self.clear
    and self.set_default are defined.
    The user shoult inherit from both a ParamSet and one ParamSetContainer
    and use the addContainer method in the inizializer.
    '''

    def _decorate(self, widget, buttons=True):
        '''Build the decoration.

        Build the decoration and pack the "widget" parameter inside it.
        If the "buttons" parameter evaluates True a button-bar with the "Clean"
        and "Dafault" buttons is added to the decoration.
        The return value is the most extern widget.
        '''

        frame = super(SubSet, self)._decorate(widget, buttons)
        self._checkbutton = gtk.CheckButton()
        self._checkbutton.set_active(True)
        self._checkbutton.connect('toggled', self.on_toggled)
        label = frame.get_label_widget()
        label.set_padding(0, 2)
        hbox = gtk.HBox(spacing=5)
        hbox.set_border_width(5)
        frame.set_label_widget(hbox)
        hbox.pack_start(self._checkbutton, expand=False)
        hbox.pack_start(label)
        tooltips = gtk.Tooltips()
        tooltips.set_tip(self._checkbutton,
            'Click on th checkbutton to enable/disable the paramset')

        return frame

    def get_enabled(self):
        return self._checkbutton.get_active()

    def set_enabled(self, enabled):
        self._checkbutton.set_active(enabled)

    enabled = property(get_enabled, set_enabled)

    def on_toggled(self, widget):
        paramset_widget = self._frame.get_child()
        paramset_widget.set_sensitive(widget.get_active())

class OptSubSet(BaseSubSet):
    '''Decoration for optional sub-set of parameters.

    This class is designed to provide decorations to ParamSet classes.
    The decoration is a frame with a bold title enclosed in an expander:
    the entire subset can be disanled by clicking on the expander arrow.
    OptSubSet is thought to be used as MixIn class together with a ParamSet
    class and it expects that some attribures/methods like self.label,
    self.clear and self.set_default are defined.
    The user shoult inherit from both a ParamSet and one ParamSetContainer
    and use the addContainer method in the inizializer.
    '''

    def _decorate(self, widget, buttons=True):
        '''Build the decoration.

        Build the decoration and pack the "widget" parameter inside it.
        If the "buttons" parameter evaluates True a button-bar with the "Clean"
        and "Dafault" buttons is added to the decoration.
        The return value is the most extern widget.
        '''

        widget = super(OptSubSet, self)._decorate(widget, buttons)
        self._expander = gtk.Expander()
        self._expander.set_use_markup(True)
        self._expander.set_label('Disabled')
        self._expander.connect('activate', self.on_expand)
        self._expander.add(widget)

        return self._expander

    def get_enabled(self):
        return self._expander.get_expanded()

    def set_enabled(self, enabled):
        if enabled:
            self._expander.set_label('')
        else:
            self._expander.set_label('Disabled')

        self._expander.set_expanded(enabled)

    enabled = property(get_enabled, set_enabled)

    def on_expand(self, widget=None):
        # NOTE: the activate signal is emitter before the "expanded"
        # property is updated
        if self.enabled:
            self._expander.set_label('Disabled')
        else:
            self._expander.set_label(None)

def decorated_paramset(paramset_type, decoration_type=None, name=None):
    '''Return a new ParamSet class with decorations.

    Both the type of the ParamSet and decoration type are passed as parameters.
    If the "decoration_type" is None then the "paramset_type" is returned.

    Example:

    >>> newParamSet = decorated_paramset(StdParamSet, MainSet)

    '''

    if decoration_type is None:
        return paramset_type

    class DecoratedParamSet(paramset_type, decoration_type):
        def __init__(self, *args, **kwargs):
            buttons = kwargs.pop('buttons', None)
            paramset_type.__init__(self, *args, **kwargs)
            assert self.widget is not None
            self._paramset_widget = self.widget
            if buttons is None:
                self.widget = self._decorate(self.widget)
            else:
                self.widget = self._decorate(self.widget, buttons)

        if hasattr(decoration_type, 'enabled'):
            enabled = decoration_type.enabled

    if not name:
        name = '%s_%s' % (paramset_type.__name__, decoration_type.__name__)
    DecoratedParamSet.__name__ = name

    return DecoratedParamSet

# -----------------------------------------------------------------------------
# --- Base param set classes --------------------------------------------------
# -----------------------------------------------------------------------------
class BaseParamSet(object):
    '''Abstract base class for sets of parameters.

    Derived classes should at least implement (and overwrite) the following
    methods: set_default, clear, set_from_cfg, update_cfg.

    The ParamSet classes implementations should never make no assumption about
    the "widget" attribute but that it is a widget.
    If it is needed to reference some widget component for methods
    implemetation it is strongly suggested to bind that widget to a (pivate)
    attribute.

    A ParamSet has at least the following attributes:

    name -- the ParamSet name
    label -- a label for the PatamSet
    widget -- the widget that encloses the ParamSet GUI

    The "widget" attribute is used for packing the ParamSet into into the
    application GUI.
    '''

    def __init__(self, name, label='', widget=None, **kwargs):
        '''ParamSet initializer.

        name -- the ParamSet name
        label -- a label for the PatamSet
        widget -- the widget that encloses the ParamSet GUI

        The name, label and widget parameters are stored as attributes.
        '''
        # NOTE: the call to the object.__init__ method is very important
        super(BaseParamSet, self).__init__()

        self.name = name
        self.label = label
        self.widget = widget
        self._workdir = kwargs.get('workdir')

    def set_default(self):
        '''Set at the default value all the objects in the ParamSet.'''

        raise NotImplementedError('%s is a virtual class.' %
                                                self.__class__.__name__)

    def clear(self):
        '''Clear all the objects in the ParamSet.'''

        raise NotImplementedError('%s is a virtual class.' %
                                                self.__class__.__name__)

    # @TODO: set values method
    #~ def set_values(self, values):
        #~ '''Set the objects in the ParamSet using the "values" parameter.

        #~ "values" can be a dictionary like object or a sequence of instance
        #~ whose class inherits from BaseParam or BaseParamSet.'''
        #~ pass

    def set_from_cfg(self, cfg, section=None, errors_='strict'):
        '''Set the objects in the ParamSet getting the values from "cfg".

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        raise NotImplementedError('%s is a virtual class.' %
                                                self.__class__.__name__)

    def update_cfg(self, cfg, section=None):
        '''Update the "cfg" object setting the values from the ParamSet.

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        raise NotImplementedError('%s is a virtual class.' %
                                                self.__class__.__name__)

    # @TODO: section (??)
    #        include this in set_from_cfg (??)
    def load(self, inifilename, section=None):
        assert inifilename.strip(), ('Empty inifilename argument passed to '
                                     '%s.save method.' %
                                                    self.__class__.__name__)

        cfg = ConfigParser.SafeConfigParser()
        cfg.read(inifilename)
        self.set_from_cfg(cfg)

    def save(self, inifilename):
        cfg = ConfigParser.SafeConfigParser()
        try:
            self.update_cfg(cfg)
        except (ConfigParser.DuplicateSectionError,
                errors.MissingParameterError,
                errors.InvalidParameterError), e:
            dialogs.run_error_dialog(str(e))
            return False

        assert inifilename.strip(), ('Empty inifilename argument passed to '
                                     '%s.save method.' %
                                                    self.__class__.__name__)

        cfg.write(file(inifilename, 'w'))
        # @TODO: handle Duplicate section, write failed
        #        (IOException, OSException)
        return True

    def get_enabled(self):
        return self.widget.get_property('sensitive')

    def set_enabled(self, sensitive):
        self.widget.set_sensitive(sensitive)

    enabled = property(get_enabled, set_enabled, None,
                       'Tell whenever the ParamSet is enabled or not.')

class StdParamSet(BaseParamSet):
    '''Standard set of parameters.

    StdParamSet implements the BaseParamSet interface. The "params" arrtibute
    is used to store the parameters. All parameters objects in "params" must
    be instances of BaseParam or BaseParamSet.
    The instatioation of parameter objects and packing into the container
    "widget" is left to the user.
    '''

    def __init__(self, params, name, label='', widget=None, **kwargs):
        '''StdParamSet initializer.

        Parameters:
            params -- a list of parameter objects implementing the BaseParam
                      interface
            name -- the name of the set
            label -- a label
            widget -- the container widget in wich params widgets are packed

        The GUI is built by the _build_GUI private method in order to allow
        to overwtite it.
        NOTE: the widget patameter, if passed, is expected to be a gtk.Frame.
        '''

        self.params = list(params)
        if widget is None:
            widget = self._build_GUI(self.params)
        super(StdParamSet, self).__init__(name, label, widget, **kwargs)

    def _build_GUI(self, params):
        '''Build the GUI.'''

        self._table = gtk.Table(len(params), 2)
        self._table.set_border_width(8)

        for indx, par in enumerate(params):
            # build the label widget
            label_widget = gtk.Label(par.label)
            label_widget.set_name('%sLabel' % par.name)
            label_widget.set_alignment(1, 0)
            try:
                if par.mandatory:
                    label_widget.set_markup('<b>%s</b>' % par.label)
            except AttributeError:
                # @TODO: improve the ParamSet handling
                pass

            # put the widgets into the table
            self._table.attach(label_widget, 0, 1, indx, indx + 1,
                               xoptions=gtk.FILL, yoptions=gtk.FILL,
                               xpadding=10, ypadding=6)
            self._table.attach(par.widget, 1, 2, indx, indx + 1,
                               yoptions=gtk.FILL)

            # finally set the label widgets as instance attributes
            setattr(self, label_widget.get_name(), label_widget)

        return self._table

    def set_default(self):
        '''Set at the default value all the objects in the ParamSet.'''

        for par in self.params:
            try:
                par.value = par.default
            except AttributeError:
                # @TODO: improve the exceptions handling.
                #        in case of failure raise the right exception type
                par.set_default()
                # @TODO: special case --> fix
                if isinstance(par, AoIParamSet):
                    par.enabled = False

    def clear(self):
        '''Clear all the objects in the ParamSet.'''

        for par in self.params:
            par.clear()
            # @TODO: special case --> fix
            if isinstance(par, AoIParamSet):
                par.enabled = False

    def set_from_cfg(self, cfg, section=None, errors_='strict'):
        '''Set the objects in the ParamSet getting the values from "cfg".

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.

        'strict'
        'warn'
        'ignore'
        '''

        if section is None:
            section = self.label
        # @TODO: allow rollback
        self.clear()
        for par in self.params:
            try:
                # assumes isinstance(par, BaseParamSet)
                par.set_from_cfg(cfg, section)
            except ConfigParser.NoSectionError, e:
                par.set_from_cfg(cfg, par.label)
            except AttributeError, e:
                if 'set_from_cfg' in str(e):
                    # it is not a param-set
                    label = par.label
                    if cfg.has_option(section, label):
                        val = cfg.get(section, label)
                        val = [utils.remove_quote(item)
                                                for item in val.split(',')]
                        val = ', '.join(val)
                        try:
                            par.value = val
                        except ValueError:
                            if errors_ == 'strict':
                                raise
                            elif errors_ == 'warn':
                                msg = ('Invalid value "%s" for the "%s" '
                                       'parameter.\nIgnoreg.') % (val, par.name)
                                dialog = gtk.MessageDialog(
                                    #self.widget.get_toplevel(),
                                    flags=gtk.DIALOG_MODAL |
                                          gtk.DIALOG_DESTROY_WITH_PARENT,
                                    type=gtk.MESSAGE_WARNING,
                                    buttons=gtk.BUTTONS_CLOSE,
                                    message_format=msg)
                                dialog.run()
                                dialog.destroy()
                else:
                    raise

    def update_cfg(self, cfg, section=None):
        '''Update the "cfg" object setting the values from the ParamSet.

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        if section is None:
            section = self.label
        if not cfg.has_section(section):
            cfg.add_section(section)
        for par in self.params:
            if not par.enabled:
                if par.label == 'Acknowledge Mount':
                    cfg.set(section, 'Acknowledge Mount', "'N'")
                continue
            try:
                value = par.value
                if value:
                    # @TODO: check shoul raise an exception with error
                    #        description.
                    if par.check() == False:
                        msg = ('Check failed for parameter "%s" (label="%s", '
                               'type="%s").\nPlease fix it and try again.' %
                                (par.name, par.label, par.__class__.__name__))
                        raise errors.InvalidParameterError(msg)
                    cfg.set(section, par.label, value)
                elif par.mandatory:
                    msg = ('Missing mandatory parameter "%s".\n'
                           'Please fill the "%s" field and try again.' %
                                (par.name, par.label))
                    raise errors.MissingParameterError(msg)
            except AttributeError:
                par.update_cfg(cfg, section)

    def __getattr__(self, name):
        for param in self.params:
            if param.name == name:
                return param
        raise AttributeError(name)

StdParamSubSet = decorated_paramset(StdParamSet, BaseSubSet, 'StdParamSubSet')
StdParamChkSubSet = decorated_paramset(StdParamSet, SubSet, 'StdParamChkSubSet')

class StdMultiParamSet(BaseParamSet):
    def __init__(self, paramsets, name, label='', workdir=None):
        '''
        paramsets -- a sequence of ParamSet instances
        '''

        super(StdMultiParamSet, self).__init__(name, label, widget=None,
                                               workdir=workdir)
        self.paramsets = list(paramsets)
        self.widget = self._build_GUI(self.paramsets)

    def _build_GUI(self, paramsets):
        self._notebook = gtk.Notebook()
        #self._notebook.set_border_width(5)
        for pset in paramsets:
            pset.widget.set_border_width(10)
            label = gtk.Label(pset.label)
            label.set_padding(6, 0)
            self._notebook.append_page(pset.widget, label)
        # NOTE: set_current_page(0) doesn't work without show_all
        self._notebook.show_all()
        self._notebook.set_current_page(0)
        return self._notebook

    def set_default(self):
        '''Set at the default value all the objects in the ParamSet.'''

        for pset in self.paramsets:
            pset.set_default()

    def clear(self):
        '''Clean all the objects in the ParamSet.'''

        for pset in self.paramsets:
            pset.clear()
            # @TODO: special case --> fix
            if isinstance(pset, AoIParamSet):
                pset.enabled = False

    def set_from_cfg(self, cfg, section=None, errors_='strict'):
        '''Set the objects in the ParamSet getting the values from "cfg".

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        if section is None:
            section = self.label
        # @TODO: allow rollback
        self.clear()
        for pset in self.paramsets:
            pset.set_from_cfg(cfg, section, errors_)

    def update_cfg(self, cfg, section=None):
        '''Update the "cfg" object setting the values from the ParamSet.

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        if section is None:
            section = self.label
        if not cfg.has_section(section):
            cfg.add_section(section)
        for pset in self.paramsets:
            if not pset.enabled:
                continue
            pset.update_cfg(cfg, section)

def _new_dirgetter(dirpar, workdir):
    def dirgetter():
        if workdir is None:
            default = ''
        if callable(workdir):
            default = utils.remove_quote(workdir())
        else:
            default = workdir

        dir_ = dirpar.get_value()
        dir_ = utils.remove_quote(dir_)
        if dir_:
            if not os.path.isabs(dir_):
                dir_ = os.path.join(default, dir_)
                dir_ = os.path.normpath(dir_)
            return dir_
        else:
            return default
    return dirgetter

class BaseCfgMixIn(object):
    def _load_params_def(self, cfg, workdir=None):
        params = []
        for section in cfg.sections():
            if section in ('MAIN', 'MENU'):
                continue
            dict_ = dict()
            for key, value in cfg.items(section):
                dict_[key] = utils.fromstring(value)
            position = int(dict_.pop('position'))
            if dict_.has_key('width'):
                dict_['width'] = int(dict_['width'])
            if dict_.has_key('action'):
                dict_['action'] = getattr(gtk, dict_['action'])

            ptype = getattr(paramtype, dict_.pop('type'))
            try:
                if issubclass(ptype, paramtype.BaseFileParam):
                    dict_.setdefault('workdir', workdir)
                params.append((position, ptype(**dict_)))
            except TypeError, e:
                if 'keyword' in str(e):
                    e.message = '%s (paramtype.%s)' % (e.message, ptype.__name__)
                    e.args = (e.message,)
                raise
        params.sort()
        params = [par for dummy, par in params]
        return params

    def _fix_workdir(self, params, workdir=''):
        params = list(params)

        # Search for inputDir/outputDir
        dirpar = None
        for par in params:
            for key in ('input',):# 'output'):
                #TODO: use par.action also
                if key in par.name.lower():
                    # @TODO: fix coding style in hmifile
                    name = '%sDir' % key
                    for p in params:
                        #TODO: use par.action also
                        if p.name == name:
                            dirpar = p
                            break

        if dirpar:
            dirgetter = _new_dirgetter(dirpar, workdir)

            # Scan for paramtype.File prameters
            for par in params:
                if isinstance(par, paramtype.File): #, paramtype.Path):
                    par.set_workdir_getter(dirgetter)
        return params

class CfgParamSet(StdParamSet, BaseCfgMixIn):
    '''ParamSet defined from configuration file.

    The format of the configuration file must be readable using the
    ConfigParser module.'''

    def __init__(self, hmifile, name='', label=None, workdir=None):

        if not hasattr(hmifile, 'sections'):
            # Load the hmi file
            assert os.path.exists(hmifile), ('"%s" configuration file does '
                                             'not exist.' % hmifile)
            cfg = ConfigParser.SafeConfigParser()
            cfg.read(hmifile)
        else:
            cfg = hmifile

        # Set the label
        if label is None:
            if cfg.has_option('MAIN', 'title'):
                label = cfg.get('MAIN', 'title')
            else:
                label = ''

        # Set the AoI flag
        if cfg.has_option('MAIN', 'has_aoi'):
            has_aoi = utils.fromstring(cfg.get('MAIN', 'has_aoi'))
            assert has_aoi in (True, False), ('Invalid value for the "has_aoi"'
                                              ' parameter.')
        else:
            has_aoi = False

        params = self._load_params_def(cfg, workdir)
        params = self._fix_workdir(params, workdir)
        if has_aoi:
            # @TODO: special case --> fix it
            par = OptAoIParamSet()
            params.append(par)

        # GUI setup
        widget = self._build_GUI(params)

        # Call the base class initializer
        super(CfgParamSet, self).__init__(params, name, label, widget,
                                          workdir=workdir)

MainCfgParamSet = decorated_paramset(CfgParamSet, MainSet, 'MainCfgParamSet')

class CfgMultiParamSet(StdMultiParamSet, BaseCfgMixIn):
    def __init__(self, hmifile, name='', label=None, workdir=None):

        if not hasattr(hmifile, 'sections'):
            # Load the hmi file
            assert os.path.exists(hmifile), ('"%s" configuration file does '
                                             'not exist.' % hmifile)
            cfg = ConfigParser.SafeConfigParser()
            cfg.read(hmifile)
        else:
            cfg = hmifile

        # Set the label
        if label is None:
            if cfg.has_option('MAIN', 'title'):
                label = cfg.get('MAIN', 'title')
            else:
                label = ''

        # Set the AoI flag
        if cfg.has_option('MAIN', 'has_aoi'):
            has_aoi = utils.fromstring(cfg.get('MAIN', 'has_aoi'))
            assert has_aoi in (True, False), ('Invalid value for the "has_aoi"'
                                              ' parameter.')
        else:
            has_aoi = False

        params = self._load_params_def(cfg, workdir)
        params = self._fix_workdir(params, workdir)
        paramsets = self._get_paramsets(params)
        if has_aoi:
            pset = AoIParamSubSet()
            paramsets.append(pset)

        super(CfgMultiParamSet, self).__init__(paramsets, name, label,
                                               workdir=workdir)

    def _get_paramsets(self, params):
        raise NotImplementedError('%s is a virtual class.' %
                                                self.__class__.__name__)

MainCfgMultiParamSet = decorated_paramset(CfgMultiParamSet, MainSet,
                                          'MainCfgMultiParamSet')

# -----------------------------------------------------------------------------
# --- Custom param sets -------------------------------------------------------
# -----------------------------------------------------------------------------
class AoIParamSet(aoi.AoIForm, BaseParamSet):
    '''Set of parameters representing the Area of Interest (AoI).

        @TODO: AoIParamSet should not derive from AoIForm, it "has" a AoIForm
               attribute.
    '''

    def __init__(self, name='AoI', label='Area of Interest'):
        aoi.AoIForm.__init__(self)
        BaseParamSet.__init__(self, name, label, self.widget)

    def set_default(self):
        '''Set at the default value all the objects in the ParamSet.'''

        aoi.AoIForm.set_default(self)

    def clear(self):
        '''Clear all the objects in the ParamSet.'''

        aoi.AoIForm.clear(self)

    def set_from_cfg(self, cfg, section=None, errors_='strict'):
        '''Set the objects in the AoIParamSet getting the values from "cfg".

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        labels = [label.lower() for label in aoi.name_to_label.values()]
        opts = set(cfg.options(section))
        if opts.intersection(labels):
            self.enabled = True
            aoi.AoIForm.set_from_cfg(self, cfg, section, errors_)
        else:
            self.enabled = False

    def update_cfg(self, cfg, section=None):
        '''Update the "cfg" object setting the values from the AoIParamSet.

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        if self.enabled:
            values = self.get_values()
            for key, value in values:
                cfg.set(section, key, value)

AoIParamSubSet = decorated_paramset(AoIParamSet, SubSet, 'AoIParamSubSet')
OptAoIParamSet = decorated_paramset(AoIParamSet, OptSubSet, 'OptAoIParamSet')

class HeaderAnalysisParamSet(MainCfgParamSet):
    # @TODO:
    #   - sensorMode is mandatory for ASAR
    #   - handling of patterns for the ERS products -> directories
    #     callback to change action when the ERS pattern is set in the file
    #     dialog
    #   - if CDROM is selected for the "Input Media Type" parameter then the
    #     default directory for the "Input Media Path" file dialog should be
    #     the cdrom

    def __init__(self, workdir=None):
        hmifile = utils.get_hmifile('header.ini')
        super(HeaderAnalysisParamSet, self).__init__(hmifile,
                                                     'HeaderAnalysisParamSet',
                                                     workdir=workdir)

        self.on_media_type_toggled()
        # @TODO: fix coding style in header.ini
        for button in self.inputMediaType.group:
            button.connect('toggled', self.on_media_type_toggled)

        self.on_sensor_id_toggled()
        # @TODO: fix coding style in header.ini
        for button in self.sensorId.group:
            button.connect('toggled', self.on_sensor_id_toggled)

        self.on_sensor_mode_toggled()
        # @TODO: fix coding style in header.ini
        for button in self.sensorMode.group:
            button.connect('toggled', self.on_sensor_mode_toggled)

        self.on_number_of_volumes_toggled()
        # @TODO: fix coding style in header.ini
        for button in self.numberOfVolumes.group:
            button.connect('toggled', self.on_number_of_volumes_toggled)

        # @TODO: fix coding style in header.ini
        self.dataFormat.enabled = False

        # Tooltips
        tooltips = gtk.Tooltips()
        # @TODO: fix coding style in header.ini
        tooltips.set_tip(self.productType.PRI, 'Precision products: IMP, APP')
        tooltips.set_tip(self.productType.SLC, 'Complex products: IMS, APS')
        tooltips.set_tip(self.productType.GEC, 'Geocoded products: IMG, APG')
        tooltips.set_tip(self.productType.MR,
            'Medium Resolution products: IMM, WSM, ...')
        tooltips.set_tip(self.productType.BRW,
            'Browse products: IM__BP, AP__BP, ...')
        tooltips.set_tip(self.productType.RAW, 'Raw data')

        tooltips.set_tip(self.sourceId.esp,
            'All Envisat data and ERS data processed at ESRIN products')
        tooltips.set_tip(self.sourceId.itp, 'ERS data processed at I-PAF')
        tooltips.set_tip(self.sourceId.ukp, 'ERS data processed at UK-PAF')
        tooltips.set_tip(self.sourceId.dep, 'ERS data processed at D-PAF')
        tooltips.set_tip(self.sourceId.fst,
            'ERS data processed by an ACS w-k processor in '
            'Argentina (Cordoba), China (Beijing), Ecuador (Cotopaxi), '
            'Israel (Tel-Aviv), Kenya (Malindi), Russia, South Africa or '
            'Thailand (Bangkok)')
        tooltips.set_tip(self.sourceId.sis,
            'ERS data processed at Singapore Station')

    def on_media_type_toggled(self, widget=None):
        value = utils.remove_quote(self.inputMediaType.value)
        enabled = value in ('CDROM', 'TAPE')
        # @TODO: fix coding style in header.ini
        self.ackMount.enabled = enabled
        self.ackMountLabel.set_sensitive(enabled)
        self.dismountVolume.enabled = enabled
        self.dismountVolumeLabel.set_sensitive(enabled)
        if enabled:
            self.on_number_of_volumes_toggled()

    def on_sensor_id_toggled(self, widget=None):
        # @TODO: fix coding style in header.ini
        if self.sensorId.ASAR.get_active():
            self.sensorMode.enabled = True
            self.on_sensor_mode_toggled()

            self.dataFormat.default = 'MPHSPH'
            self.dataFormat.value = 'MPHSPH'

            self.sourceId.value = 'esp'
            self.sourceId.enabled = False

            self.inputMediaPath.action = gtk.FILE_CHOOSER_ACTION_OPEN

            #self.numberOfVolumes.enabled = False # @TODO: check
        else:
            self.sensorMode.enabled = False
            self.APDataset.enabled = False

            self.dataFormat.default = 'CEOS'
            self.dataFormat.value = 'CEOS'

            self.sourceId.enabled = True

            self.inputMediaPath.action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER

            #self.numberOfVolumes.enabled = True # @TODO: check

    def on_sensor_mode_toggled(self, widget=None):
        value = utils.remove_quote(self.sensorMode.value)
        if value == 'Alternating Polarization':
            self.APDataset.enabled = True
        else:
            self.APDataset.enabled = False

    def on_number_of_volumes_toggled(self, widget=None):
        if (self.dismountVolume.enabled and
            int(self.numberOfVolumes.value) > 1):
            self.dismountVolume.value = 'Y'

    def update_cfg(self, cfg, section=None):
        super(HeaderAnalysisParamSet, self).update_cfg(cfg, section)
        if section is None:
            section = self.label
        cfg.set(section, self.dataFormat.label, self.dataFormat.value)
        if not self.sourceId.enabled:
            cfg.set(section, self.sourceId.label, self.sourceId.value)

class QuickLookParamSet(MainCfgParamSet):
    # @TODO:
    #   - handling of patterns for the ERS products -> directories
    #   - callback to change action when the ERS pattern is set in the file
    #     dialog.

    def __init__(self, workdir=None):
        hmifile = utils.get_hmifile('quick.ini')
        super(QuickLookParamSet, self).__init__(hmifile, 'QuickLookParamSet',
                                                workdir=workdir)

        # @TODO: fix coding style in quick.ini
        dirgetter = _new_dirgetter(self.inputDir, workdir)
        self.headerAnalysisFile.set_workdir_getter(dirgetter)

        self.on_media_type_toggled()
        # @TODO: fix coding style in quick.ini
        for button in self.inputMediaType.group:
            button.connect('toggled', self.on_media_type_toggled)

    def on_media_type_toggled(self, widget=None):
        # @TODO: fix coding style in quick.ini
        value = utils.remove_quote(self.inputMediaType.value)

        enabled = (value != 'FILE')
        # @TODO: fix coding style in quick.ini
        self.headerAnalysisFile.enabled = enabled
        self.headerAnalysisFileLabel.set_sensitive(enabled)

        enabled = value in ('TAPE', 'CDROM')
        self.ackMount.enabled = enabled
        self.ackMountLabel.set_sensitive(enabled)
        self.dismountVolume.enabled = enabled
        self.dismountVolumeLabel.set_sensitive(enabled)

        if value == 'FILE':
            self.inputMediaPath.action = gtk.FILE_CHOOSER_ACTION_OPEN
        else:
            self.inputMediaPath.action = None

class FullResolutionParamSet(MainCfgParamSet):

    def __init__(self, workdir=None):
        hmifile = utils.get_hmifile('full.ini')
        super(FullResolutionParamSet, self).__init__(hmifile,
                                                     'FullResolutionParamSet',
                                                     workdir=workdir)

        # @TODO: fix coding style in full.ini
        dirgetter = _new_dirgetter(self.inputDir, workdir)
        self.headerAnalysisFile.set_workdir_getter(dirgetter)

        self.on_media_type_toggled()
        for button in self.inputMediaType.group:
            button.connect('toggled', self.on_media_type_toggled)

    def on_media_type_toggled(self, widget=None):
        value = utils.remove_quote(self.inputMediaType.value)

        enabled = value in ('TAPE', 'CDROM')
        self.ackMount.enabled = enabled
        self.ackMountLabel.set_sensitive(enabled)
        self.dismountVolume.enabled = enabled
        self.dismountVolumeLabel.set_sensitive(enabled)

        if value == 'FILE':
            self.inputMediaPath.action = gtk.FILE_CHOOSER_ACTION_OPEN
        else:
            self.inputMediaPath.action = None

#~ class ImportRasterParamSet(CfgParamSet):
    #~ # @TODO:
    #~ #   - tooltips
    #~ pass

class GCPSelectionForm(gladewrap.GladeObject):
    name_to_label = {
        'gcpNumbers': 'GCPs Numbers',
        'gcpFile': 'GCPs File Name',
    }

    def __init__(self):
        gladefile = utils.get_gladefile('bestgui.glade')
        super(GCPSelectionForm, self).__init__(gladefile, 'gcp_selection_frame')
        self.combobox.set_active(0)

        # GCPs File
        patterns = (
            ('*.txt *.dat', 'ASCII files (.txt .dat)'),
            ('*.txt', 'Text files (.txt)'),
            ('*.dat', 'ASCII data files (.dat)'),
        )
        self.gcp_filename = paramtype.File('gcpFile',
                                           self.name_to_label['gcpFile'],
                                           patterns=patterns)
        self.gcpfile_hbox.pack_start(self.gcp_filename.widget)

    def set_default(self):
        '''Set at the default value all the objects in the ParamSet.'''

        # "GCP's Numbers"
        self.combobox.set_active(0)
        self.row_spinbutton.set_value(15)
        self.col_spinbutton.set_value(15)

    def clear(self):
        '''Clean all the objects in the ParamSet.'''

        page = self.combobox.get_active()
        if page == 0:
            # "GCP's Numbers"
            self.row_spinbutton.set_value(0)
            self.col_spinbutton.set_value(0)
        elif page == 1:
            # "GCP File Name" # page = 1
            self.gcp_filename.clear()

    def get_values(self):
        values = []
        page = self.combobox.get_active()
        if page == 0:
            # "GCP's Numbers"
            row = self.row_spinbutton.get_value_as_int()
            col = self.col_spinbutton.get_value_as_int()
            label = self.name_to_label['gcpNumbers']
            values.append((label, '%d, %d' % (row, col)))
        elif page == 1:
            # "GCP File Name" # page = 1
            label = self.name_to_label['gcpFile']
            values.append((label, self.gcp_filename.value))

        return values

    def set_values(self, values):
        values = dict(values)

        label = self.name_to_label['gcpNumbers'].lower()
        if values.has_key(label):
            row, col = values[label].split(',')
            row = int(row.strip())
            col = int(col.strip())
            self.row_spinbutton.set_value(row)
            self.col_spinbutton.set_value(col)
            index = self.name_to_label.keys().index('gcpNumbers')
            self.combobox.set_active(index)

        label = self.name_to_label['gcpFile'].lower()
        if values.has_key(label):
            self.gcp_filename.value = utils.remove_quote(values[label])
            index = self.name_to_label.keys().index('gcpFile')
            self.combobox.set_active(index)

    def on_combobox_changed(self, widget=None):
        index = widget.get_active()
        self.notebook.set_current_page(index)

    def on_edit(self, widget=None):
        dialog = aoi.VertexEditorDialog()
        dialog.set_coorsys_enabled(False)
        dialog.filename = utils.remove_quote(self.gcp_filename.value)
        dialog.load(dialog.filename)
        dialog.run()
        workdir = self.gcp_filename.workdir
        self.gcp_filename.value = utils.fix_path(workdir, dialog.filename)
        dialog.destroy()

class GCPSelectionParamSet(GCPSelectionForm, BaseParamSet):
    def __init__(self, name='gcp_selection', label='GCP Selection'):
        BaseParamSet.__init__(self, name, label)
        super(GCPSelectionParamSet, self).__init__() # set the widget attribute
        self.enabled = False

    def set_default(self):
        '''Set at the default value all the objects in the ParamSet.'''

        GCPSelectionForm.set_default(self)

    def clear(self):
        '''Clean all the objects in the ParamSet.'''

        GCPSelectionForm.clear(self)

    def set_from_cfg(self, cfg, section=None, errors_='strict'):
        '''Set the objects in the AoIParamSet getting the values from "cfg".

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        labels = [label.lower() for label in self.name_to_label.values()]
        opts = set(cfg.options(section))
        if opts.intersection(labels):
            self.enabled = True
            values = {}
            for label in labels:
                if cfg.has_option(section, label):
                    try:
                        values[label] = cfg.get(section, label)
                    except ValueError:
                        if errors_ == 'strict':
                            raise
                        elif errors_ == 'warn':
                            msg = ('Invalid value "%s" for the "%s" '
                                   'parameter.\nIgnoreg.' % (values[label],
                                                                        label))
                            dialog = gtk.MessageDialog(
                                #self.widget.get_toplevel(),
                                flags=(gtk.DIALOG_MODAL |
                                      gtk.DIALOG_DESTROY_WITH_PARENT),
                                type=gtk.MESSAGE_WARNING,
                                buttons=gtk.BUTTONS_CLOSE,
                                message_format=msg)
                            dialog.run()
                            dialog.destroy()
            self.set_values(values)
        else:
            self.enabled = False

    def update_cfg(self, cfg, section=None):
        '''Update the "cfg" object setting the values from the AoIParamSet.

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        values = self.get_values()
        for key, value in values:
            cfg.set(section, key, value)

    def get_enabed(self):
        return self.combobox.get_active() != 2

    def set_enabled(self, enabled):
        if enabled:
            self.combobox.set_active(0)
        else:
            self.combobox.set_active(2)

    enabled = property(get_enabed, set_enabled)

class AltitudeOfAmbiguityForm(gladewrap.GladeObject):

    name_to_label = {
        'altAmbiguityEval':     'Altitude of Ambiguity Evaluation',
        'inCoorType':           'Input Coordinates Type',
        'numberOfGridPoints':   'Number of Points',
        'latLonPoints':         'Input Coordinates',
        'rowColPoints':         'Input Coordinates',
    }

    def __init__(self):
        gladefile = utils.get_gladefile('bestgui.glade')
        super(AltitudeOfAmbiguityForm, self).__init__(
                    gladefile, 'altitude_of_ambiguity_frame')

        self.checkString = paramtype.CheckString('altAmbiguityEval',
                            ':EVALUATE', self.name_to_label['altAmbiguityEval'])
        # @TODO: remove reference to private attribute
        self.checkString._checkbutton.connect('toggled', self.on_toggled)
        self.hbox.pack_start(self.checkString.widget)

        self.vertexeditor = aoi.VertexEditor()
        self.notebook.append_page(self.vertexeditor.widget,
                                  gtk.Label('POINTS'))

        self.combobox.set_active(0)
        self.combobox.set_sensitive(False)
        self.notebook.set_current_page(0)
        self.notebook.set_sensitive(False)

    def set_default(self):
        '''Set at the default value all the objects in the ParamSet.'''

        # POINTGRID
        self.combobox.set_active(0)
        self.row_spinbutton.set_value(15)
        self.col_spinbutton.set_value(15)

    def clear(self):
        '''Clear all the objects in the ParamSet.'''

        page = self.combobox.get_active()
        if page == 0:
            # POINTGRID
            self.row_spinbutton.set_value(0)
            self.col_spinbutton.set_value(0)
        else:
            # POINTS
            self.vertexeditor.on_clear_all_items()

    def get_values(self):
        vlist = []

        val = self.checkString.value
        if val:
            vlist.append((self.name_to_label['altAmbiguityEval'], val))

        page = self.combobox.get_active()
        if page == 0:
            # POINTGRID
            vlist.append((self.name_to_label['inCoorType'], 'POINTGRID'))
            row = self.row_spinbutton.get_value_as_int()
            col = self.col_spinbutton.get_value_as_int()
            key = self.name_to_label['numberOfGridPoints']
            vlist.append((key, '%d, %d' % (row, col)))
        else:
            # POINTS
            values = self.vertexeditor.get_values()
            values = dict(values)
            key = aoi.name_to_label['coordinateSystem']
            vlist.append((self.name_to_label['inCoorType'], values[key]))
            key = aoi.name_to_label['vertexList']
            vlist.append(('Input Coordinates', values[key]))
        return vlist

    def set_values(self, vlist):
        dict_ = dict(vlist)

        # @TODO: fix coding style
        key = self.name_to_label['numberOfGridPoints'].lower()
        value = dict_.get(key)
        if value:
            row, col = value.split(',')
            row = int(row.strip())
            col = int(col.strip())
            self.row_spinbutton.set_value(row)
            self.col_spinbutton.set_value(col)
            self.combobox.set_active(0)

        key = self.name_to_label['rowColPoints'].lower()
        value = dict_.get(key)
        if not value:
            key = self.name_to_label['latLonPoints'].lower()
            value = dict_.get(key)
        if value:
            value = [(aoi.name_to_label['vertexList'].lower(), value)]
            self.vertexeditor.set_values(value)
            self.combobox.set_active(1)

        key = self.name_to_label['inCoorType'].lower()
        coor_type = dict_.get(key)
        if coor_type == 'POINTGRID':
            self.combobox.set_active(0)
        elif coor_type in self.vertexeditor.coorSysLabelsMap.keys():
            self.combobox.set_active(1)
            self.vertexeditor.set_coorsys(coor_type)

    def on_coortype_changed(self, widget=None):
        index = widget.get_active()
        self.notebook.set_current_page(index)

    def on_toggled(self, widget=None):
        value = utils.remove_quote(self.checkString.value)
        if value == self.checkString.options[1]:
            # enable
            self.combobox.set_sensitive(True)
            self.notebook.set_sensitive(True)
        else:
            # disable
            self.combobox.set_sensitive(False)
            self.notebook.set_sensitive(False)

class AltitudeOfAmbiguityParamSet(AltitudeOfAmbiguityForm, BaseParamSet):
    def __init__(self, name='altitudeOfAmbiguity',
                 label='Altitude of Ambbiguity Evaluation'):
        BaseParamSet.__init__(self, name, label)
        # set the widget attribute
        super(AltitudeOfAmbiguityParamSet, self).__init__()

    def set_default(self):
        '''Set at the default value all the objects in the ParamSet.'''

        AltitudeOfAmbiguityForm.set_default(self)
        self.enabled = False

    def clear(self):
        '''Clear all the objects in the ParamSet.'''

        AltitudeOfAmbiguityForm.clear(self)

    def set_from_cfg(self, cfg, section=None, errors_='strict'):
        '''Set the objects in the AoIParamSet getting the values from "cfg".

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        labels = [label.lower() for label in self.name_to_label.values()]
        opts = set(cfg.options(section))
        if opts.intersection(labels):
            self.enabled = True
            values = {}
            for label in labels:
                if cfg.has_option(section, label):
                    try:
                        values[label] = cfg.get(section, label)
                    except ValueError:
                        if errors_ == 'strict':
                            raise
                        elif errors_ == 'warn':
                            msg = ('Invalid value "%s" for the "%s" '
                                   'parameter.\nIgnoreg.' % (values[label],
                                                                        label))
                            dialog = gtk.MessageDialog(
                                #self.widget.get_toplevel(),
                                flags=(gtk.DIALOG_MODAL |
                                       gtk.DIALOG_DESTROY_WITH_PARENT),
                                type=gtk.MESSAGE_WARNING,
                                buttons=gtk.BUTTONS_CLOSE,
                                message_format=msg)
                            dialog.run()
                            dialog.destroy()
            self.set_values(values)
        else:
            self.enabled = False

    def update_cfg(self, cfg, section=None):
        '''Update the "cfg" object setting the values from the AoIParamSet.

        "cfg" must be a "ConfigParser" object (from the python standard
        library) and the section parameter is used to specify the configuration
        file section. If "section" is None then the "label" attribute is used
        as section specifierqualifier.
        '''

        if self.enabled:
            values = self.get_values()
            for key, value in values:
                cfg.set(section, key, value)
        else:
            value = self.checkString.value
            if value:
                cfg.set(section, self.name_to_label['altAmbiguityEval'], value)

    def get_enabled(self):
        value = utils.remove_quote(self.checkString.value)
        return value == self.checkString.options[1]

    def set_enabled(self, enabled):
        enabled = bool(enabled)
        if self.enabled != enabled:
            self.checkString._checkbutton.set_active(enabled)

    enabled = property(get_enabled, set_enabled)

class CoRegistrationParamSet(MainCfgMultiParamSet):

    # @TODO: fix codinf style in coreg.ini
    main_params_names = [
        'inputDir',         # "Input Dir"                       pos = 0  OK
        'inputImages',      # "Input Images"                    pos = 1  OK
        'outputDir',        # "Output Dir"                      pos = 2  OK
        'outputImages',     # "Output Images"                   pos = 3  OK

        'delInputImage',    # "Delete Input Image"              pos = 4  OK

        'overMode',         # "Overlapping Mode"                pos = 25 OK
        'minOverAOI',       # "Overlapping AoI Threshold"       pos = 27 OK

        'baselineFile',     # "Baseline File Name"              pos = 5  OK
        'resFileName',      # "Residual File Name"              pos = 26 OK
    ]

    gcp_params_names = [
        'gcpNumbers',       # "GCPs Numbers"                    pos = 6  OK
        'gcpFile',          # "GCPs File Name"                  pos = 7  OK
        'editingRMS',       # "Editing RMS"                     pos = 8  OK
    ]

    coarse_reg_params_names = [
        'coarseRegFlag',    # "Image Coarse Reg"                pos = 9  OK
        'coarseCellDim',    # "Coarse Reg Window Sizes"         pos = 10 OK
        'coarseInterpFactor', # "Coarse Reg Interp Factors"     pos = 11 OK
        'coarseTol',        # "Coarse Reg Tolerance"            pos = 12 OK
    ]

    fine_reg_params_names = [
        'fineRegFlag',      # "Image Fine Reg"                  pos = 13 OK
        'fileCellDim',      # "Fine Reg Window Sizes"           pos = 14 OK
        'csoil',            # "Coherence Threshold"             pos = 15 OK
        'coherWinSize',     # "Coherence Window Size"           pos = 16 OK
        'coherFTol',        # "Coherence Func Tolerance"        pos = 17 OK
        'coherVTol',        # "Coherence Value Tolerance"       pos = 18 OK
    ]

    interpolation_params_names = [
        'warpDegree',       # "Transformation Degree"           pos = 19 OK
        'interWinSizes',    # "Interp Window Sizes"             pos = 20 OK
        'interMode',        # "Interpolation Mode"              pos = 21 OK
        'interPrecis',      # "Interpolation Inverse Precision" pos = 22 OK
        'sincWidth',        # "Sinc Width"                      pos = 23 OK
        'ccCoeff',          # "Cubic Convolution Coefficient"   pos = 24 OK
    ]

    extra_output_params_names = [
        'baselineEvaluation',  # "Baseline Evaluation"              pos = 28
        'altAmbiguityEval',    # "Altitude of Ambiguity Evaluation" pos = 29
        'inCoorType',          # "Input Coordinates Type"           pos = 30
        'numberOfGridPoints',  # "Number of Points"                 pos = 31
        'latLonPoints',        # "Input Coordinates"                pos = 32
        'rowColPoints',        # "Input Coordinates"                pos = 33

        'residualEvaluation',  # "Residual Evaluation"              pos = 34
        'qualityEvaluation',   # "Quality Evaluation"               pos = 35
        'coherenceEvaluation', # "Coherence Evaluation"             pos = 36
    ]

    subsets = (
        ('Main', main_params_names),
        ('GCP', gcp_params_names),
        ('Coarse Reg', coarse_reg_params_names),
        ('Fine Reg', fine_reg_params_names),
        ('Interpolation', interpolation_params_names),
        ('Extra Output', extra_output_params_names),
    )

    def __init__(self, workdir=None):
        hmifile = utils.get_hmifile('coreg.ini')
        super(CoRegistrationParamSet, self).__init__(hmifile,
                                                     'coregistration_paramset',
                                                     workdir=workdir)

        subset_labels = [label for label, dummy in self.subsets]

        # @TODO: fix coding style in hmi file
        coarse_reg_index = subset_labels.index('Coarse Reg')
        fine_reg_index = subset_labels.index('Fine Reg')
        interpolation_index = subset_labels.index('Interpolation')

        index_name_callback = (
            (coarse_reg_index, 'coarseRegFlag', self.on_subset_toggled),
            (fine_reg_index, 'fineRegFlag', self.on_subset_toggled),
            (interpolation_index, 'interMode', self.on_interp_mode_changed),
        )

        for index, name, callback in index_name_callback:
            paramset = self.paramsets[index]
            par_names = [par.name for par in paramset.params]
            par_index = par_names.index(name)
            param = paramset.params[par_index]
            for button in param.group:
                button.connect('toggled', callback, index, par_index)
            param.widget.show_all()
            callback(param.group[0], index, par_index)

        aoi_paramset = self.paramsets[-1]
        aoi_paramset.enabled = False

    def _get_paramsets(self, params):
        params = dict([(par.name, par) for par in params])
        paramsets = []
        for label, names in self.subsets:
            pset = []
            for name in names:
                pset.append(params.pop(name))
            name = label.replace(' ', '')
            if label == 'GCP':
                # fix the patamset (I know it is a StdParamSet)
                params_to_remove = [par for par in pset if par.label in
                                    GCPSelectionParamSet.name_to_label.values()]
                for par in params_to_remove:
                    pset.remove(par)
                    par.widget.destroy()

                # add the GCPSelectionParamSet
                gcp_selection = GCPSelectionParamSet()
                pset.insert(0, gcp_selection)

            elif label == 'Extra Output':
                # fix the patamset (I know it is a StdParamSet)
                # @TODO: remove wired labels
                labels_to_remove = (
                    'Altitude of Ambiguity Evaluation',
                    'Input Coordinates Type',
                    'Number of Points',
                    'Input Coordinates'
                )
                params_to_remove = [par for par in pset
                                            if par.label in labels_to_remove]
                for par in params_to_remove:
                    pset.remove(par)
                    par.widget.destroy()

                # add the GCPSelectionParamSet
                altitude_of_ambiguity_pset = AltitudeOfAmbiguityParamSet()
                pset.insert(1, altitude_of_ambiguity_pset)

            pset = StdParamSubSet(pset, name, label)
            paramsets.append(pset)

        if params:
            warnings.warn('The HMI file defines unused parameters that will '
                'be ignored: %s' % ', '.join([p.name for p in params]))

        return paramsets

    # @TODO: check
    #~ def update_cfg(self, cfg, section=None):
        #~ # @TODO: remove wired labels
        #~ super(CoRegistrationParamSet, self).update_cfg(cfg, section)
        #~ if section is None:
            #~ section = self.label
        #~ if not cfg.has_option(section, 'Number of Points'):
            #~ cfg.remove_option(section, 'Input Coordinates Type')

    def on_subset_toggled(self, widget, pset_index, par_index):
        label = widget.get_label()
        if (((label == 'Y') and widget.get_active()) or
            ((label == 'N') and not widget.get_active())):
            enabled = True
        else:
            enabled = False

        paramset = self.paramsets[pset_index]
        for index, par in enumerate(paramset.params):
            if index == par_index:
                continue
            par.widget.set_sensitive(enabled)

    def on_interp_mode_changed(self, widget=None, pset_index=None,
                               par_index=None):
        # @TODO: remove wired labels
        paramset = self.paramsets[pset_index]
        par = paramset.params[par_index]
        value = utils.remove_quote(par.value)

        par_names = [par.name for par in paramset.params]

        sinc_width = paramset.params[par_names.index('sincWidth')]
        sinc_width.enabled = ('SINC' in value)

        cubic_convolution_coeff = paramset.params[par_names.index('ccCoeff')]
        cubic_convolution_coeff.enabled = ('CUBIC' in value)

if(__name__ == '__main__'):
    def test_ParamSet():
        win = gtk.Window()
        win.connect('destroy', gtk.main_quit)
        win.set_default_size(800, 600)
        hmi = CfgParamSet('../config/hmi/quick.ini')
        win.add(hmi.widget)
        win.show_all()
        gtk.main()

    def test_enabled():
        widget = gtk.Label()
        for ParamSet in (BaseParamSet, StdParamSet):
            print '%s:' % ParamSet.__name__,
            pset = ParamSet(widget=widget)
            print 'enabled = %s' % pset.enabled

        print '%s:' % CfgParamSet.__name__,
        pset = CfgParamSet('../config/hmi/quick.ini')
        print 'enabled = %s' % pset.enabled

        print '%s:' % AoIParamSet.__name__,
        pset = AoIParamSet()
        print 'enabled = %s' % pset.enabled

    def test_HeaderAnalysisParamSet():
        win = gtk.Window()
        win.connect('destroy', gtk.main_quit)
        win.set_default_size(1000, 800)
        hmi = HeaderAnalysisParamSet()
        win.add(hmi.widget)
        win.show_all()
        gtk.main()

    def test_QuickLookParamSet():
        win = gtk.Window()
        win.connect('destroy', gtk.main_quit)
        win.set_default_size(1000, 800)
        hmi = QuickLookParamSet()
        win.add(hmi.widget)
        win.show_all()
        gtk.main()

    def test_FullResolutionParamSet():
        win = gtk.Window()
        win.connect('destroy', gtk.main_quit)
        win.set_default_size(1000, 800)
        hmi = FullResolutionParamSet()
        win.add(hmi.widget)
        win.show_all()
        gtk.main()

    def test_CoRegistrationParamSet():
        win = gtk.Window()
        win.connect('destroy', gtk.main_quit)
        win.set_default_size(800, 600)
        hmi = CoRegistrationParamSet()
        win.add(hmi.widget)
        win.show_all()
        gtk.main()

    #test_ParamSet()
    #test_enabled()
    #test_HeaderAnalysisParamSet()
    #test_QuickLookParamSet()
    #~ test_FullResolutionParamSet()
    test_CoRegistrationParamSet()
