### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''Utility functions and classes for parameter insertion handling.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__version__ = '$Revision: 704 $'
__date__    = '$Date: 2009-04-13 10:09:18 +0200 (lun, 13 apr 2009) $'

import os

import gtk

import utils
import errors
import dialogs
import filepatterns

def checknum(value, condition):
    if callable(condition):
        return condition(value)
    if isinstance(condition, basestring):
        try:
            condition = float(condition)
            return value == condition
        except ValueError:
            if 'VALUE' in condition:
                return utils.safe_eval(condition, dict(range=range),
                                                            dict(VALUE=value))
            else:
                return utils.safe_eval('%s %s' % (str(value), condition),
                                                            dict(range=range))
    else:
        return value == condition

class BaseParam(object):
    def __init__(self, name, label='', mandatory=False, default=None):
        self.name      = name
        self.mandatory = mandatory
        self.default   = default
        self.label     = label
        self.widget    = None
        super(BaseParam, self).__init__() # @NOTE: very important

    def clear(self):
        self.set_value(None)

    def check(self):
        return(True)

    def __str__(self):
        return '%s = %s' % (self.name, self.value)

    def get_value(self):
        if self.__class__.__name__ == 'BaseParam':
            raise NotImplementedError('%s is a virtual class.' %
                                      self.__class__.__name__)
        else:
            return self.get_value()

    def set_value(self, value):
        if self.__class__.__name__ == 'BaseParam':
            raise NotImplementedError('%s is a virtual class.' %
                                      self.__class__.__name__)
        else:
            self.set_value(value)

    value = property(get_value, set_value, None, 'Parameter value.')

    def get_enabled(self):
        return self.widget.get_property('sensitive')

    def set_enabled(self, enabled):
        self.widget.set_sensitive(enabled)

    enabled = property(get_enabled, set_enabled)

class Entry(BaseParam):
    '''Base class for parameters that require an entry widget.'''
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=10, description=''):
        super(Entry, self).__init__(name, label, mandatory, default)

        self._entry = gtk.Entry()
        self._entry.set_width_chars(width)
        self._entry.connect('populate-popup', self.on_populate_popup)
        self.widget = gtk.HBox()
        self.widget.set_spacing(5)
        self.widget.set_name(name)
        self.widget.pack_start(self._entry)#, expand=True, fill=False)
        if description:
            description_label = gtk.Label(description)
            self.widget.pack_start(description_label, expand=False, padding=1)

    def get_value(self):
        return(self._entry.get_text())

    def set_value(self, value):
        if value is None:
            self._entry.set_text('')
        else:
            self._entry.set_text(str(value))

    def on_clear(self, widget, data=None):
        self.clear()

    def on_set_default(self, widget, data=None):
        self.value = self.default

    def on_populate_popup(self, widget, menu, data=None):
        separator = gtk.SeparatorMenuItem()
        separator.show()
        menu.append(separator)

        # Clean
        #item = gtk.MenuItem('_Clean')
        item = gtk.ImageMenuItem(gtk.STOCK_CLEAR)
        item.connect('activate', self.on_clear, None)
        item.show()
        menu.append(item)

        # Default
        item = gtk.MenuItem('_Default')
        item.connect('activate', self.on_set_default)
        item.show()
        menu.append(item)

        return False

    #~ def __getattr__(self, name):
        #~ return getattr(self._entry, name) # @TODO: remove this

class String(Entry):
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=40, description=''):
        super(String, self).__init__(name, label, mandatory, default, width,
                                     description)

    def get_value(self):
        val = super(String, self).get_value()
        if val:
            return('"%s"' % val)
        return('')

class Radio(BaseParam):
    '''Base class for parameters that require radio buttons'''

    def __init__(self, name, options, label='', mandatory=False, default=None):
        super(Radio, self).__init__(name, label, mandatory, default)

        # Check options
        if isinstance(options, basestring):
            options = options.split(':')
        if not options:
            msg = ('Invalid option set: "%s".\n'
                   'A colon separated list of values expected.' % options)
            raise ValueError(msg)

        self.group = None

        self.widget = gtk.HBox()
        self.widget.set_spacing(5)
        self.widget.set_name(name)
        self.set_options(options)

        self.value = self.default

    def get_value(self):
        for button in self.group:
            if button.get_active():
                return(button.get_label())

    def set_value(self, value):
        if value is None:
            return
        for button in self.group:
            if button.get_label().lower() == value.lower():
                button.set_active(True)
                return
        options = ', '.join([button.get_label() for button in self.group])
        msg = ('Invalid value "%s" for %s. '
               'Allowed values are %s' % (value, self.name, options))
        raise ValueError(msg)

    def get_options(self):
        return [button.get_label() for button in self.group]

    def set_options(self, options):
        if len(options) < 2:
            raise ValueError('At least two options expected.')
        # @TODO: check
        if len(options) * max(map(len, options)) > 100:
            container = gtk.VBox()
            container.set_spacing(5)
            # @TODO: do not refer widget. Use a private attribute instead
            self.widget.pack_start(container)
        else:
            # @TODO: do not refer widget. Use a private attribute instead
            container = self.widget
        radiobutton = None
        for item in options:
            item = item.strip()
            radiobutton = gtk.RadioButton(radiobutton, item)
            radiobutton.set_name(item)
            container.pack_start(radiobutton, expand=False, fill=False)
            radiobutton.show()
        self.group = radiobutton.get_group()

    options = property(get_options, set_options, None,
                       'The radio button options.')

    def __getattr__(self, name):
        try:
            return self.group[self.options.index(name)]
        except ValueError:
            raise AttributeError(name)

class RadioScalar(Radio):
    pass

class RadioChar(Radio):
    def get_value(self):
        for button in self.group:
            if button.get_active():
                return("'%s'" % button.get_label())

class RadioString(Radio):
    '''Base class for parameters that require radio buttons

    The "NONE" string is treated as a special value.

    '''

    def get_value(self):
        for button in self.group:
            if button.get_active():
                value = button.get_label()
                if value.lower() == 'none':
                    return ''
                else:
                    return('"%s"' % value)

class Check(BaseParam):
    '''Base class for parameters that require check buttons.

    options -- a sequence of lenght two or a colon separated list of two values

    The first item in "options" (the one with index 0) is the value of the
    parameter when the button is non checked, the second (the one with index 1)
    is the value of the parameter when the button is checked.

    '''

    def __init__(self, name, options, label='', mandatory=False, default=None):
        super(Check, self).__init__(name, label, mandatory, default)

        # Check options
        if isinstance(options, basestring):
            options = options.split(':')
        if not options or len(options) != 2:
            msg = ('Invalid option set: "%s".\n'
                   'A colon separated list of two values or a sequence of '
                   'lenght two is expected.' % options)
            raise ValueError(msg)

        self._checkbutton = gtk.CheckButton()
        self._checkbutton.connect('toggled', self.on_toggled)
        self._checkbutton.set_name(name)
        self.options = list(options)
        if self.default is None:
            self.default = options[0]

        if self.default == '':
            if not self.options[0]:
                label = self.options[1]
            else:
                label = self.options[0]
            self._checkbutton.set_label(label)

        self.value = self.default
        self.widget = self._checkbutton

    def get_value(self):
        if self._checkbutton.get_active():
            index = 1
        else:
            index = 0
        return self.options[index]

    def set_value(self, value):
        if value is None:
            return
        if value not in self.options:
            options = ', '.join(self.options)
            msg = ('Invalid value "%s" for %s. '
                   'Allowed values are %s' % (value, self.name, options))
            raise ValueError(msg)

        if utils.remove_quote(self.get_value()) != value:
            index = self.options.index(value)
            self._checkbutton.set_active(index)

    def on_toggled(self, widget=None, data=None):
        if '' not in self.options:
            val = utils.remove_quote(self.get_value())
            if val == self.options[0]:
                label = self.options[1]
            else:
                label = self.options[0]
            self._checkbutton.set_label(label)

class CheckChar(Check):
    def get_value(self):
        val = super(CheckChar, self).get_value()
        if val:
            return("'%s'" % val)
        return('')

class CheckString(Check):
    def get_value(self):
        val = super(CheckString, self).get_value()
        if val:
            return('"%s"' % val)
        return('')

#~ class Range(BaseParam):
    #~ pass

class Scalar(Entry):
    # @TODO: range redefines a builtin type
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=6, range=None):
        self.range = range
        super(Scalar, self).__init__(name, label, mandatory, default, width,
              description=range)

    def check(self):
        if self.range:
            # @TODO: fix --> update hmi files
            range = self.range
            if ' - ' in range:
                min_, max_ = map(float, range.split(' - '))
                def fun(value):
                    return min_ <= value <= max_
                range = fun
            checknum(self.value, range)
        else:
            return True

    #range = property() # @TODO:

class Vector(Entry):
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=10, description='', lencheck=None):
        super(Vector, self).__init__(name, label, mandatory, default, width,
                                     description)
        self.lencheck = lencheck

    def check(self):
        if self.lencheck:
            # @TODO: check
            return checknum(len(self.value.split(',')), self.lencheck)
        else:
            return True

class StringVector(Vector):
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=40, description='', lencheck=None):
        super(StringVector, self).__init__(name, label, mandatory, default,
                                           width, description, lencheck)

    def get_value(self):
        result = []
        s = super(StringVector, self).get_value()
        for val in s.split(','):
            val = utils.remove_quote(val) # perform a preliminar strip
            result.append('"%s"' % val)
        if result:
            return ', '.join(result)
        else:
            return('')

class BaseFileParam(Entry):
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=40, multiple=False, numcheck=None, workdir=None):
        # @TODO: move the button definition here
        super(BaseFileParam, self).__init__(name, label, mandatory, default,
                                            width)
        # @COMPATYBILITY: conversion required by PyGTK <= 2.8.6
        self.multiple = int(multiple)
        self.numcheck = numcheck
        self._workdir = workdir

    def get_value(self):
        result = []
        s = super(BaseFileParam, self).get_value()
        for val in s.split(','):
            val = utils.remove_quote(val) # perform a preliminar strip
            if val:
                val = '"%s"' % val
            result.append(val)
        return ', '.join(result)

    def set_value(self, value):
        if value is None:
            self._entry.set_text('')
        else:
            parts = [utils.remove_quote(val) for val in value.split(',')]
            self._entry.set_text(', '.join(parts))

    def get_workdir(self):
        if self._workdir is None:
            return ''
        if callable(self._workdir):
            return utils.remove_quote(self._workdir())
        else:
            return self._workdir

    def set_workdir_getter(self, workdir):
        assert ((workdir is None) or callable(workdir) or
                 isinstance(workdir, basestring)), ('"workdir" must be a '
                                    'string, a callable object or None')
        self._workdir = workdir

    workdir = property(get_workdir)

    def check(self):
        result = super(BaseFileParam, self).check()
        if self.numcheck:
            n = len(self.value.split(','))
            return result and checknum(n, self.numcheck)
        return result

class File(BaseFileParam):
    # @TODO: force extension for FILE_CHOOSER_ACTION_SAVE
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=40, action=None, patterns=[], multiple=False,
                 numcheck=None, workdir=None):
        super(File, self).__init__(name, label, mandatory, default, width,
                                   multiple, numcheck, workdir)
        if not action:
            if 'out' in name:
                self.action = gtk.FILE_CHOOSER_ACTION_SAVE
            else:
                self.action = gtk.FILE_CHOOSER_ACTION_OPEN

        self.patterns = self.parse_patterns(patterns)

        # GUI setup
        if self.action == gtk.FILE_CHOOSER_ACTION_SAVE:
            button = gtk.Button(stock=gtk.STOCK_SAVE_AS)
        else:
            button = gtk.Button(stock=gtk.STOCK_OPEN)

        button.connect('clicked', self.on_button_clicked)
        button.show()
        self.widget.pack_start(button, expand=False, fill=False)

    def parse_patterns(self, patterns):
        if not patterns:
            return patterns
        if isinstance(patterns, basestring):
            # @TODO: check security
            patterns = utils.safe_eval(patterns)
            if patterns:
                if isinstance(patterns[0], basestring):
                    patterns = [patterns]
                else:
                    patterns = list(patterns)
        return patterns

    def set_patterns(self, filechooser):
        if not self.patterns:
            return
        patterns = list(self.patterns)
        patterns = filepatterns.find_patterns(patterns)
        patterns.append(('*', 'All Files'))
        for pattern, name in patterns:
            filter_ = gtk.FileFilter()
            filter_.set_name(name)
            for pattern in pattern.split():
                filter_.add_pattern(pattern)
            filechooser.add_filter(filter_)

    def on_button_clicked(self, widget=None, data=None):
        # @TODO: add a new attribute for the dialog instance and move the
        #        dialog setup to the __init__ method
        chooser_dialog = gtk.FileChooserDialog(
                            title   = self.label,
                            parent  = self.widget.get_toplevel(),
                            action  = self.action,
                            buttons = (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        self.set_patterns(chooser_dialog)

        # @TODO: pass the working directory function as a data?
        workdir = utils.remove_quote(self.workdir)

        inifilename = utils.remove_quote(self.value)
        if not os.path.isabs(inifilename):
            inifilename = os.path.join(workdir, inifilename)
        path, name = os.path.split(inifilename)

        if os.path.isabs(path):
            chooser_dialog.set_current_folder(path)
            if name:
                chooser_dialog.set_filename(inifilename)
        elif name:
            chooser_dialog.set_current_name(name)
        else:
            chooser_dialog.set_current_folder(workdir)

        chooser_dialog.set_select_multiple(self.multiple)
        chooser_dialog.set_default_response(gtk.RESPONSE_OK)
        if self.action == gtk.FILE_CHOOSER_ACTION_SAVE:
            filename = dialogs.run_save_dialog(chooser_dialog)
            if filename:
                if path:
                    filename = utils.fix_path(path, filename)
                self.set_value(filename)
        else:
            response = chooser_dialog.run()
            if response == gtk.RESPONSE_OK:
                filenames = chooser_dialog.get_filenames()
                if workdir:
                    filenames = [utils.fix_path(workdir, name)
                                                    for name in filenames]
                self.set_value(', '.join(filenames))
        chooser_dialog.destroy()

class Dir(BaseFileParam):
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=40, multiple=False, numcheck=None, workdir=None):
        super(Dir, self).__init__(name, label, mandatory, default, width,
                                  multiple, numcheck, workdir)

        # GUI setup
        self.action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER
        button = gtk.Button(stock=gtk.STOCK_OPEN)
        button.connect('clicked', self.on_button_clicked)
        button.show()
        self.widget.pack_start(button, expand=False, fill=False)

    def get_value(self):
        result = []
        s = super(Dir, self).get_value()
        for val in s.split(','):
            val = utils.remove_quote(val) # perform a preliminar strip
            if val:
                if not val.endswith(os.path.sep):
                    val = val + os.path.sep
                val = '"%s"' % val
            result.append(val)
        return ', '.join(result)

    def on_button_clicked(self, widget=None, data=None):
        chooser_dialog = gtk.FileChooserDialog(
                            title   = self.label,
                            parent  = self.widget.get_toplevel(),
                            action  = self.action,
                            buttons = (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                                       gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        path = utils.remove_quote(self.value)
        if not path:
            path = self.workdir
        elif not os.path.isabs(path):
            path = os.path.join(self.workdir, path)
        chooser_dialog.set_current_folder(path)

        chooser_dialog.set_select_multiple(self.multiple)
        chooser_dialog.set_default_response(gtk.RESPONSE_OK)
        answer = chooser_dialog.run()
        if(answer == gtk.RESPONSE_OK):
            dirnames = chooser_dialog.get_filenames()
            if self.workdir:
                fixed_dirs = []
                for dir_ in dirnames:
                    dir_ = utils.fix_path(self.workdir, dir_)
                    if not os.path.isabs(dir_):
                        dir_ = os.path.join('.', dir_)
                    fixed_dirs.append(dir_)
                dirnames = fixed_dirs
                # @TODO: fix only if dirname is not an abspath
                #dirnames = [os.path.join('.', name) for name in dirnames]
            self.set_value(', '.join(dirnames))
        chooser_dialog.destroy()

# NOTE: the order of the superclasses is important in this case.
#       The lookup starts afrer "File" in the method resolution order (__mro__)
#       so if "File" comes first of "Dir" then Dir.__init__ is called after
#       File.__init__ and tho buttons are created.
class Path(Dir, File):
    # @TODO: use a custom dialog with product type (CEOS/MPHSPH) selector
    def __init__(self, name, label='', mandatory=False, default=None,
                 width=40, action=None, patterns=[], workdir=None):
        File.__init__(self, name, label, mandatory, default, width, None,
                      patterns, False, workdir=workdir)
        self.action = action

    def on_button_clicked(self, widget=None, data=None):
        if self.action in (gtk.FILE_CHOOSER_ACTION_OPEN,
                           gtk.FILE_CHOOSER_ACTION_SAVE):
            File.on_button_clicked(self, widget, data)
        elif self.action == gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER:
            Dir.on_button_clicked(self, widget, data)
        elif self.action is None:
            selector_dialog = gtk.FileSelection(title = self.label)
            target = utils.remove_quote(self.value)
            if not os.path.isabs(target):
                target = os.path.join(self.workdir, target)
            if os.path.exists(target):
                selector_dialog.set_filename(target)
                if os.path.isdir(target):
                    selector_dialog.selection_entry.set_text('')
            else:
                target = self.workdir
                if target:
                    selector_dialog.set_filename(target)
                    selector_dialog.selection_entry.set_text('')
            selector_dialog.show_fileop_buttons()
            selector_dialog.set_select_multiple(False)
            selector_dialog.set_default_response(gtk.RESPONSE_OK)

            answer = selector_dialog.run()
            if(answer == gtk.RESPONSE_OK):
                path = selector_dialog.get_filename()
                self.set_value(path)
            selector_dialog.destroy()
        else:
            raise errors-BestGUIException('Invalid actiion in %s: %s' %
                                        (self.__class__.__name__, self.action))

    def get_value(self):
        if self.action in (gtk.FILE_CHOOSER_ACTION_OPEN,
                           gtk.FILE_CHOOSER_ACTION_SAVE):
            return File.get_value(self)
        elif self.action in (gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
                             gtk.FILE_CHOOSER_ACTION_CREATE_FOLDER):
            return Dir.get_value(self)
        else:
            val = utils.remove_quote(BaseFileParam.get_value(self))
            if val:
                if os.path.isdir(val) and not val.endswith(os.path.sep):
                    val = val + os.path.sep
                return('"%s"' % val)
            return('')

if(__name__ == '__main__'):
    def test_enabled():
        for Param in (Entry, Radio, Scalar, RadioScalar, Vector, RadioChar,
                      RadioString, String, File, Dir, Path):
            print '%s:' % Param.__name__,
            par = Param('name', 'a : b')
            print 'enabled = %s' % par.enabled

    test_enabled()
