### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''Program preferences.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__date__    = '$Date: 2009-04-13 10:09:18 +0200 (lun, 13 apr 2009) $'
__version__ = '$Revision: 704 $'

import os
import sys
import logging
import ConfigParser
import webbrowser

import gtk
import pango

import utils
import errors
import dialogs

from gladewrap import GladeDialog
from errors import BestEnvError

class OptionsDialog(GladeDialog):
    # @TODO: tags settings

    def __init__(self, filename, cfg=None, parent=None):
        gladefile = utils.get_gladefile('bestgui.glade')
        super(OptionsDialog, self).__init__(gladefile, 'options_dialog')
        self.set_default_response(gtk.RESPONSE_OK)
        self.set_default(self.ok_button)

        index = self.get_level_index('INFO')
        self.loglevel_combobox.set_active(index)

        self._best_env = {
            'bestbin': '',
            'besthome': '',
            'flagfile': '',
        }
        self.filename_entry.set_text(filename)

        if not cfg:
            filename = os.path.expanduser(filename)
            if os.path.exists(filename):
                cfg = ConfigParser.RawConfigParser()
                cfg.read(filename)
        if cfg:
            self.set_options(cfg)

        if parent:
            self.set_transient_for(parent)

    def get_level_index(self, level):
        if not isinstance(level, basestring):
            level = logging.getLevelName(level)
            if level.startswith('Level'):
                raise ValueError('invalid logging level: "%s"' %level)
        model = self.loglevel_combobox.get_model()
        levels = [row[0] for row in model]
        return levels.index(level)

    def use_environment(self, use_env=True):
        self.use_env_checkbutton.set_active(use_env)

        if use_env:
            # store the current environment settings
            for option in ('bestbin', 'besthome', 'flagfile'):
                widget = getattr(self, option + '_entry')
                if widget.get_property('sensitive'):
                    self._best_env[option] = widget.get_text()

            BESTHOME = os.getenv('BESTHOME', '')
            if not BESTHOME:
                BESTHOME = os.getenv('STBXHOME', '')
                if BESTHOME:
                    self.besthome_label.set_text('STBXHOME')
            else:
                self.besthome_label.set_text('BESTHOME')
            self.besthome_entry.set_text(BESTHOME)

            FLAGFILE = os.getenv('FLAGFILE', '')
            self.flagfile_entry.set_text(FLAGFILE)

            bestbin = utils.which('best')
            if not bestbin:
                bestbin = utils.which('stbx')
            self.bestbin_entry.set_text(bestbin)
        else:
            self.besthome_label.set_text('BESTHOME')
            for name in ('bestbin', 'besthome', 'flagfile'):
                widget = getattr(self, name + '_entry')
                widget.set_text(self._best_env[name])

        sensitive = not use_env
        for name in ('bestbin', 'besthome', 'flagfile'):
            widget = getattr(self, name + '_entry')
            widget.set_sensitive(sensitive)
            widget = getattr(self, name + '_button')
            widget.set_sensitive(sensitive)

    def set_options(self, cfg):
        # BESTGUI section
        section = 'BESTGUI'
        if cfg.has_section(section):
            if cfg.has_option(section, 'show_preview'):
                value = cfg.get(section, 'show_preview')
                radiobutton = getattr(self, '%s_radiobutton' % value.lower())
                radiobutton.set_active(True)

            for option in ('workdir', 'browserlauncher'):
                if cfg.has_option(section, option):
                    value = cfg.get(section, option)
                    getattr(self, '%s_entry' % option).set_text(value)

            if cfg.has_option(section, 'loglevel'):
                loglevel = cfg.get(section, 'loglevel')
                index = self.get_level_index(loglevel)
                self.loglevel_combobox.set_active(index)

        # BEST section
        section = 'BEST'
        if cfg.has_section(section):
            for option in ('bestbin', 'besthome', 'flagfile'):
                if cfg.has_option(section, option):
                    value = cfg.get(section, option)
                    self._best_env[option] = value
                    getattr(self, '%s_entry' % option).set_text(value)

            if cfg.has_option(section, 'use_environment'):
                value = cfg.get(section, 'use_environment')
                use_env = utils.fromstring(value)
                self.use_environment(use_env)

    def get_options(self):
        cfg = ConfigParser.RawConfigParser()

        section = 'BESTGUI'
        if not cfg.has_section(section):
            cfg.add_section(section)
        for option in ('workdir', 'browserlauncher'):
            value = getattr(self, '%s_entry' % option).get_text()
            if value:
                cfg.set(section, option, value)
            else:
                cfg.remove_option(section, option)
        group = self.always_radiobutton.get_group()
        for radiobutton in group:
            if radiobutton.get_active():
                cfg.set(section, 'show_preview',
                          radiobutton.get_label().lower())
                break

        loglevel = self.loglevel_combobox.get_active_text()
        cfg.set(section, 'loglevel', loglevel)

        section = 'BEST'
        if not cfg.has_section(section):
            cfg.add_section(section)
        if self.use_env_checkbutton.get_active():
            cfg.set(section, 'use_environment', 'True')
            for option in ('bestbin', 'besthome', 'flagfile'):
                cfg.set(section, option, self._best_env[option])
        else:
            cfg.set(section, 'use_environment', 'False')
            for option in ('bestbin', 'besthome', 'flagfile'):
                value = getattr(self, '%s_entry' % option).get_text()
                if value:
                    cfg.set(section, option, value)
                else:
                    cfg.remove_option(section, option)

        return cfg

    def on_use_env_checkbutton_toggled(self, widget=None):
        active = self.use_env_checkbutton.get_active()
        self.use_environment(active)

    def on_open_button_clicked(self, widget=None):
        name = widget.get_name()
        if name:
            name = name.replace('_button', '')
        title = getattr(self, '%s_label' % name).get_text().title()
        entry = getattr(self, '%s_entry' % name)

        if name in ('besthome', 'workdir'):
            action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER
        else: #if name in ('bestbin', 'flagfile'):
            action = gtk.FILE_CHOOSER_ACTION_OPEN

        dialog = gtk.FileChooserDialog(
                    title   = title,
                    parent  = widget.get_toplevel(),
                    action  = action,
                    buttons = (gtk.STOCK_OK,     gtk.RESPONSE_OK,
                               gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))

        value = entry.get_text().strip()
        if value:
            value = os.path.expandvars(os.path.expanduser(value))
            value = os.path.normpath(value)
            dialog.set_filename(value)

        if name in ('bestbin', 'flagfile'):
            besthome = self.besthome_entry.get_text()
            if besthome:
                dialog.set_current_folder(besthome)

        dialog.set_select_multiple(False)
        dialog.set_default_response(gtk.RESPONSE_OK)
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            entry.set_text(dialog.get_filename())
        dialog.destroy()

    def on_entry_activate(self, widget=None):
        return self.activate_default()


def get_bestbin(toolname='best', env=None):
    '''Returns the absolute path name of the BEST executable if available or
    None.'''

    bestbin = utils.which(toolname, env)
    if bestbin:
        return bestbin

    if not env:
        env = os.environ

    TOOLHOME = os.path.splitext(toolname)[0].upper() + 'HOME'
    besthome = env.get(TOOLHOME)
    if not besthome:
        flagfile = env.get('FLAGFILE')
        if flagfile:
            besthome = os.path.dirname(flagfile)

    if besthome:
        bestbin = os.path.join(besthome, 'bin', toolname)
        if utils.isexecutable(bestbin):
            return bestbin
        bestbin = os.path.join(besthome, toolname)
        if utils.isexecutable(bestbin):
            return bestbin

def get_best_home(toolname='best', env=None):
    '''Returns the BESTHOME directory if available or None.'''

    if not env:
        env = os.environ

    TOOLHOME = os.path.splitext(toolname)[0].upper() + 'HOME'
    besthome = env.get(TOOLHOME)
    if not besthome:
        bestbin = get_bestbin(toolname, env)
        if bestbin:
            path = os.path.dirname(bestbin)
            if os.path.isdir(os.path.join(path, 'cfg')):
                besthome = os.path.abspath(path)
            path = os.path.join(path, os.path.pardir)
            if os.path.isdir(os.path.join(path, 'cfg')):
                besthome = os.path.abspath(path)
    if not besthome:
        flagfile = env.get('FLAGFILE')
        if flagfile:
            besthome = os.path.dirname(flagfile)

    if besthome:
        besthome = os.path.expanduser(besthome)
        besthome = os.path.expandvars(besthome)

    return besthome

def get_best_flagfile(toolname='best', env=None):
    '''Returns the BEST FLAGFILE full path name if available or None.'''

    if not env:
        env = os.environ

    flagfile = env.get('FLAGFILE')
    if not flagfile:
        besthome = get_best_home(toolname, env)
        if besthome:
            flagfile = os.path.join(besthome, 'flagfile')
            if os.path.isfile(flagfile):
                return os.path.abspath(flagfile)
            flagfile = None
    else:
        flagfile = os.path.expanduser(flagfile)
    return flagfile

class Settings(object):
    def __init__(self, filename=None):
        super(Settings, self).__init__()
        self._cfg = None
        self.filename = None

        self.loglevel = logging.INFO
        self.workdir = os.path.expanduser('~')
        self.browserlauncher = None
        self.show_preview = 'ask'
        self.tags = {
            'error': dict(foreground='red'),
            'warning': dict(foreground='orange'),
            'info': dict(foreground='blue'),
            'debug': dict(foreground='gray'),
            'cmd': dict(weight=pango.WEIGHT_BOLD),
            'completed': dict(foreground='blue'),
            #'mesage': dict(),
        }

        self.bestbin = None
        self.best_env = {
            'BESTHOME': None,
            'FLAGFILE': None,
            'PATH': os.getenv('PATH', ''),
            'STBXHOME': None,
        }

        self.window_position = None
        self.window_size = None
        self.outputplane_fraction = 0.2

        if filename:
            self.load(filename)

    def load(self, filename):
        # NOTE: the interpolation feature of SafeConfigParser messes with the
        # webbrowser command string if it contains "%s"
        cfg = ConfigParser.RawConfigParser()
        cfg.read(os.path.expanduser(filename))
        self._cfg = cfg
        self.filename = filename

        self.update(self._cfg)

    def get_cfg(self):
        self._update_cfg()
        return self._cfg

    def _update_bestgui_options(self):

        section = 'BESTGUI'
        if not self._cfg.has_section(section):
            self._cfg.add_section(section)

        # Set loglevel
        loglevel = None
        if self._cfg.has_option(section, 'loglevel'):
            loglevel = self._cfg.get(section, 'loglevel')
            if not loglevel or loglevel == 'DEFAULT':
                loglevel = logging.INFO
            else:
                try:
                    loglevel = getattr(logging, loglevel)
                except AttributeError:
                    logging.warn('invalid log level: "%s"' % loglevel)
                    loglevel = None
        if loglevel:
            self.loglevel = loglevel
        else:
            self.loglevel = logging.INFO

        # Set workdir
        workdir = None
        if self._cfg.has_option(section, 'workdir'):
            workdir = self._cfg.get(section, 'workdir')
            workdir = utils.remove_quote(workdir)
            workdir = os.path.expanduser(workdir)
            workdir = os.path.expandvars(workdir)
        # @TODO: use stat here
        if workdir and not os.access(workdir, os.X_OK):
            msg = ("'%s' does'n exist or you don't have the correct "
                   "permissions.\n"
                   "Working directory not set." % workdir)
            logging.warning(msg)
            workdir = None
        if workdir:
            self.workdir = workdir
        else:
            self.workdir = '.'

        # Set browserlauncher
        browserlauncher = None
        if self._cfg.has_option(section, 'browserlauncher'):
            browsername = self._cfg.get(section, 'browserlauncher')
            if browsername:
                try:
                    browserlauncher = webbrowser.get(browsername)
                except TypeError:
                    # @COMPATIBILITY
                    # NOTE: in python 2.4 webbrowser.get('firefox') raises a
                    #       TypeError
                    if sys.version_info[:3] < (2, 5, 0):
                        logging.info('unable to get a webbrowser '
                                     'launcher "%s", '
                                     'use the default one' % browserlauncher)
                    else:
                        raise
                except webbrowser.Error:
                    logging.info('unable to get a webbrowser launcher "%s", '
                                 'use the default one' % browserlauncher)
                browserlauncher = None
        if browserlauncher:
            self.browserlauncher = browserlauncher
        else:
            self.browserlauncher = webbrowser.get()

        # Set show_preview
        show_preview = None
        if self._cfg.has_option(section, 'show_preview'):
            show_preview = self._cfg.get('BESTGUI', 'show_preview')
            show_preview = utils.fromstring(show_preview)
        if show_preview:
            self.show_preview = show_preview
        else:
            self.show_preview = 'ask'

        # Set the window state
        if self._cfg.has_option(section, 'window_position'):
            pos = self._cfg.get(section, 'window_position')
            try:
                pos = utils.safe_eval(pos)
                assert len(pos) == 2
            except (AssertionError, TypeError, ValueError):
                logging.warning('ignore invalid position: "%s"' % pos)
            else:
                self.window_position = pos

        if self._cfg.has_option(section, 'window_size'):
            size = self._cfg.get(section, 'window_size')
            try:
                size = utils.safe_eval(size)
                assert len(size) == 2
            except (AssertionError, TypeError):
                logging.warning('ignore invalid size: "%s"' % size)
            else:
                self.window_size = size

        if self._cfg.has_option(section, 'outputplane_fraction'):
            self.outputplane_fraction = self._cfg.getfloat(section,
                                                        'outputplane_fraction')

        # Tags
        for tag in self.tags.keys():
            props = dict()
            #props = self.tags[tag] # @TODO: check
            for tag_property in ('color', 'weight', 'style'):
                option_name = '%s_tag_%s' % (tag, tag_property)
                if self._cfg.has_option(section, option_name):
                    option_value = self._cfg.get(section, option_name)
                    if not option_value:
                        props[tag_property] = None
                    elif tag_property == 'color':
                        tag_property = 'foreground'
                        props[tag_property] = option_value
                    else:
                        try:
                            option_value = getattr(pango, option_value)
                            props[tag_property] = option_value
                        except AttributeError:
                            logging.warn('invalid value for text tag property '
                                         '"%s": %s' % (tag_property,
                                                       option_value))
            if props:
                self.tags[tag] = props

    def _update_best_options(self):
        section = 'BEST'
        if not self._cfg.has_section(section):
            self._cfg.add_section(section)

        # Set the use_environment flag
        if self._cfg.has_option(section, 'use_environment'):
            use_env = self._cfg.get(section, 'use_environment')
            use_env = utils.fromstring(use_env)
        else:
            use_env = False

        # Reset the environment
        self.bestbin = None
        self.best_env = {
                'BESTHOME': None,
                'FLAGFILE': None,
                'PATH': os.getenv('PATH', ''),
                'STBXHOME': None,
        }

        # Set the best environment (best_env) and bestbin
        if not use_env:
            # NOTE: variables expansion is performed in _fix_best_env
            if self._cfg.has_option(section, 'bestbin'):
                self.bestbin = self._cfg.get(section, 'bestbin')
            if self._cfg.has_option(section, 'besthome'):
                self.best_env['BESTHOME'] = self._cfg.get(section, 'besthome')
            if self._cfg.has_option(section, 'flagfile'):
                self.best_env['FLAGFILE'] = self._cfg.get(section, 'flagfile')
            self.best_env['STBXHOME'] = self.best_env['BESTHOME']

        for key in ('BESTHOME', 'FLAGFILE', 'STBXHOME'):
            if not self.best_env.get(key):
                self.best_env[key] = os.getenv(key)

        toolnames = ('best', 'stbx')
        if use_env:
            if os.getenv('STBXHOME') and not os.getenv('BESTHOME'):
                toolnames = tuple(reversed(toolnames))
        else:
            if self.bestbin and 'stbx' in os.path.basename(self.bestbin).lower():
                toolnames = tuple(reversed(toolnames))
        TOOLHOMES = [toolname.upper() + 'HOME' for toolname in toolnames]
        try:
            self._fix_best_env(toolnames[0])
            self.best_env[TOOLHOMES[1]] = self.best_env[TOOLHOMES[0]]
        except errors.BestEnvError, e:
            try:
                self._fix_best_env(toolnames[1])
                self.best_env[TOOLHOMES[0]] = self.best_env[TOOLHOMES[1]]
            except errors.BestEnvError:
                raise errors.BestEnvError(*e.args)


    def update(self, cfg):
        old_cfg = self._cfg
        self._cfg = cfg
        try:
            self._update_bestgui_options()
            self._update_best_options()
        except: # @TODO: check
            self._cfg = old_cfg
            raise

    def _update_cfg(self):
        if not self._cfg:
            self._cfg = ConfigParser.RawConfigParser()

            section = 'BESTGUI'
            self._cfg.add_section(section)

            for option in ('workdir', 'show_preview'):
                value = getattr(self, option)
                if value:
                    self._cfg.set(section, option, value)

            loglevel = self.loglevel
            if not isinstance(loglevel, basestring):
                loglevel = logging.getLevelName(loglevel)
            self._cfg.set(section, 'loglevel', loglevel)

            if self.browserlauncher:
                self._cfg.set(section, 'browserlauncher',
                                                    self.browserlauncher.name)

            # @TODO: save tags

            section = 'BEST'
            self._cfg.add_section(section)
            for option in ('bestbin', 'besthome', 'flagfile'):
                value = getattr(self, option)
                if value:
                    self._cfg.set(section, option, value)

        section = 'BESTGUI'
        for option in ('show_preview', 'outputplane_fraction',
                       'window_position', 'window_size'):
            value = getattr(self, option)
            if value is not None:
                self._cfg.set(section, option, str(value))

    def save(self, filename=None):
        if not filename and not self.filename:
            raise ValueError('a valid file name must be specified')
        if not filename:
            filename = self.filename

        self._update_cfg()

        try:
            self._cfg.write(file(os.path.expanduser(filename), 'w'))
        except (IOError, OSError), e:
            msg = ('Unable to save %s.\n'
                   '%s\nPreferences will be lost when you close the '
                   'program.') % (filename, str(e))
            logging.error(msg, exc_info=True)
        self.filename = filename

    def _fix_best_env(self, toolname='best'):
        '''The smart environment handling.'''

        # @TODO: variable expansion (like os.path.expanduser and
        #        os.path.expandvars)

        assert toolname in ('best', 'stbx')

        # BEST executable
        if not self.bestbin:
            self.bestbin = get_bestbin(toolname, env=self.best_env)
        if not self.bestbin:
            raise BestEnvError('Unable to locate a the %s executable.\n'
                               'Please set it manually using the Preferences '
                               'Dialog.' % toolname.upper())
        else:
            abspath = utils.which(os.path.expanduser(self.bestbin))
            if not abspath:
                raise BestEnvError("'%s' does'n exist or you don't have the "
                                   "correct permissions." % self.bestbin)

        path = os.path.dirname(self.bestbin)
        if path:
            path = os.path.expanduser(path)
            if not path in os.getenv('PATH', '').split(os.pathsep):
                path = path + os.pathsep + os.getenv('PATH', '')
                self.best_env['PATH'] = path
            #self.bestbin = os.path.basename(self.bestbin) # @TODO: check

        # BESTHOME
        TOOLHOME = '%sHOME' % toolname.upper()
        if not self.best_env[TOOLHOME]:
            self.best_env[TOOLHOME] = get_best_home(toolname, self.best_env)
        if not self.best_env[TOOLHOME]:
            raise BestEnvError('Unable to locate the "%s" directory.\n'
                               'Please set it manually using the Preferences '
                               'Dialog.' % TOOLHOME)
        else:
            self.best_env[TOOLHOME] = os.path.expanduser(
                                                    self.best_env[TOOLHOME])
            # @TODO: use stat here
            if not os.access(self.best_env[TOOLHOME], os.R_OK | os.X_OK):
                msg = ("'%s' does'n exist or you don't have the required "
                       "permissions." % self.best_env[TOOLHOME])
                if os.environ.has_key('BESTHOME'):
                    msg = '\n'.join([msg, "Plase check the '%s' "
                                          "environment variable" % TOOLHOME])
                raise BestEnvError(msg)

        # FLAGFILE
        if not self.best_env['FLAGFILE']:
            self.best_env['FLAGFILE'] = get_best_flagfile(toolname,
                                                          self.best_env)
        if not self.best_env['FLAGFILE']:
            raise BestEnvError('Unable to locate the "flagfile".\n'
                    'Please set it manually using the Preferences Dialog.')
        else:
            self.best_env['FLAGFILE'] = os.path.expanduser(
                                                    self.best_env['FLAGFILE'])
            # @TODO: use stat here
            if not os.access(self.best_env['FLAGFILE'], os.R_OK):
                msg = ("'%s' does'n exist or you don't have the required "
                       "permissions." % self.best_env['FLAGFILE'])
                if os.environ.has_key('FLAGFILE'):
                    msg = '\n'.join([msg, "Plase check the 'FLAGFILE' "
                                          "environment variable"])
                raise BestEnvError(msg)

    def set_env(self):
        '''@DEPRECATED: pass the ful executable path, workdir and env
        to the subprocess controller.'''

        # @TODO: remove this method; use subbrocess.Popen env parameter instead
        os.chdir(os.path.expanduser(self.workdir))
        if self.best_env.get('FLAGFILE'):
            os.environ['FLAGFILE'] = self.best_env['FLAGFILE']
        if self.best_env.get('BESTHOME'):
            os.environ['BESTHOME'] = self.best_env['BESTHOME']

            # This provides a partial support to "SAR TookBox" (old name of BEST).
            os.environ['STBXHOME'] = self.best_env['BESTHOME']

        if self.best_env.has_key('PATH'):
            os.environ['PATH'] = self.best_env['PATH']

if __name__ == '__main__':
    def dump_cfg(cfg):
        result = []
        for section in cfg.sections():
            result.append('[%s]' % section)
            for option, value in cfg.items(section):
                result.append('%s = %s' % (option, value))
            result.append('')
        return '\n'.join(result)

    def test_OptionDialog():
        import gobject

        def run(*args):
            try:
                filename = '~/testcfg.ini'
                d = OptionsDialog(filename)
                response = d.run()
                cfg = d.get_options()
                d.destroy()
                print dump_cfg(cfg)
                if response == gtk.RESPONSE_OK:
                    filename = os.path.expanduser(filename)
                    cfg.write(file(filename, 'w'))
            finally:
                gtk.main_quit()

        gobject.timeout_add(10, run)
        gtk.main()

    def test_Settings():
        import gobject
        import errors

        def run(*args):
            try:
                filename = '~/testcfg.ini'
                settings = Settings()
                done = False
                cfg = None
                response = None
                while not done:
                    try:
                        if not cfg:
                            settings.load(filename)
                            d = OptionsDialog(filename, cfg)
                            response = d.run()
                            cfg = d.get_options()
                            d.destroy()
                            settings.update(cfg)
                        else:
                            print dump_cfg(cfg)
                            settings.update(cfg)
                    except errors.BestEnvError, e:
                        dialogs.run_error_dialog(message=e)

                        d = OptionsDialog(filename, cfg)
                        response = d.run()
                        if response == gtk.RESPONSE_OK:
                            cfg = d.get_options()
                        else:
                            done = True
                        d.destroy()
                    else:
                        done = True
                        if response == gtk.RESPONSE_OK:
                            settings.save()
                        result = []
                        for attr_name in dir(settings):
                            attr = getattr(settings, attr_name)
                            if (not attr_name.startswith('_') and
                                not callable(attr)):
                                result.append('%s = %s' % (attr_name, attr))
                        print '\n'.join(result)

            finally:
                gtk.main_quit()

        gobject.timeout_add(10, run)
        gtk.main()
    test_Settings()

#~ pango.WEIGHT_ULTRALIGHT
#~ pango.WEIGHT_LIGHT
#~ pango.WEIGHT_NORMAL
#~ pango.WEIGHT_BOLD
#~ pango.WEIGHT_ULTRABOLD
#~ pango.WEIGHT_HEAVY

#~ pango.STYLE_NORMAL
#~ pango.STYLE_OBLIQUE
#~ pango.STYLE_ITALIC
