### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''Utility classes and functions.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__date__    = '$Date: 2009-03-14 10:51:08 +0100 (sab, 14 mar 2009) $'
__version__ = '$Revision: 701 $'

import os
import sys
import stat
import itertools
import webbrowser
import subprocess

# -----------------------------------------------------------------------------
# --- Application search paths ------------------------------------------------
# -----------------------------------------------------------------------------
try:
    from _paths import prefix as _prefix
    _sharedir   = os.path.join(_prefix, 'share', 'bestgui')
    _configdir  = _sharedir
    _docdir     = os.path.join(_prefix, 'share', 'doc', 'bestgui')
    _htmldocdir = os.path.join(_docdir, 'html')
except ImportError:
    _libDir     = os.path.abspath(os.path.dirname(__file__))
    _sharedir   = os.path.normpath(os.path.join(_libDir, os.path.pardir))
    _configdir  = os.path.join(_sharedir, 'config')
    _docdir     = os.path.join(_sharedir)
    _htmldocdir = os.path.join(_docdir, 'doc', 'html')

def get_gladefile(filename):
    '''Return the full pathname of the gladefile.'''

    return os.path.join(_sharedir, 'glade2', filename)

def get_iconfile(filename):
    '''Return the full pathname of the icon.'''

    return os.path.join(_sharedir, 'glade2/pixmaps', filename)

def get_hmifile(filename):
    '''Return the full pathname of the HMI file.'''

    return os.path.join(_configdir, 'hmi', filename)

def get_cfgfile(filename):
    '''Return the full pathname of the configuration file.'''

    return os.path.join(_configdir, filename)

def get_docfile(*args):
    '''Return the full pathname of the docfile.

    If more than one argument is passed then all the arguments are joined
    together with the base doc directory in order to form the full pathname.'''

    if os.path.splitext(args[-1])[-1] in ('.html', '.htm'):
        return os.path.join(_htmldocdir, *args)
    return os.path.join(_docdir, *args)

# -----------------------------------------------------------------------------
# --- BEST environment (@TODO: move to a separate file) -----------------------
# -----------------------------------------------------------------------------
def get_best_version(bestbin='best', env=None):
    '''Returns the BEST version string if available or None.'''

    if env:
        env.setdefault('PATH', os.getenv('PATH', ''))
    else:
        env = os.environ
    if sys.platform[:3] == 'win':
        closefds = False
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
    else:
        closefds = True
        startupinfo = None
    PIPE = subprocess.PIPE

    try:
        child = subprocess.Popen([bestbin], stdin=PIPE, stdout=PIPE,
                                 stderr=PIPE, startupinfo=startupinfo,
                                 close_fds=closefds, env=env)
        #~ child.stdin.close()
        #version_str = child.stdout.read()
        version_str, error_str = child.communicate()

        for line in version_str.splitlines():
            line = line.strip()
            if not line:
                continue
            linesplit = line.split('Generic Tool')
            if len(linesplit) > 1:
                return(linesplit[1].strip())
    except OSError:
        pass
    return None

# -----------------------------------------------------------------------------
# --- BESTGUI environment -----------------------------------------------------
# -----------------------------------------------------------------------------
def get_user_cfgfile():
    '''Return the user configuration file path.'''

    home = os.path.expanduser('~')
    if home:
        if os.name == 'nt':
            cfgfilename = os.path.join(home, 'bestgui.ini')
        else:
            cfgfilename = os.path.join(home, '.bestguirc')
        return(cfgfilename)

# -----------------------------------------------------------------------------
# --- Miscellanea -------------------------------------------------------------
# -----------------------------------------------------------------------------
def fix_browser_cmd(browsercmd):
    '''Manages the browser command (browsercmd) in order to guarantee a correct
    initialization of the webbrowser.GenericBrowser.'''

    cmd = browsercmd.strip().split()
    if len(cmd) == 1:
        cmd.append('"%s" &')
    else:
        if cmd[-1] != '&':
            cmd.append('&')
        if '%s' not in browsercmd:
            cmd.insert(-1, '"%s"')
    return ' '.join(cmd)

def remove_quote(str_):
    '''Remove <"> or <'> characters from both ends of the input string.'''

    result = str_.strip()
    for leftchr, rightchr in (('"', '"'), ("'", "'")):
        if result.startswith(leftchr) and result.endswith(rightchr):
            result = result[1:-1]
            break
    return result

def common_prefix(*sequences):
    '''Return the list of common elements at the start of all sequences, than
    a list of lists that are the unique tails of each sequence.'''

    if not sequences:
        return [], []
    common = []
    for first, second in itertools.izip(*sequences):
        if first != second:
            break
        common.append(first)
    return common, [sequence[len(common):] for sequence in sequences]

#TODO: look at the os.path.relpath functin of Python 2.6
def relative_path(path1, path2, sep=os.path.sep, pardir=os.path.pardir):
    '''Return the relative path from path1 equivalent to path2.

    In particular:
        the empty string, if path1 == path1;
        path2 if path1 and path2 have no common prefix.

    >>> pathName = os.path.join('/a','b','c')
    >>> filename = os.path.join('/c','d','e','filename.ext')
    >>> relative_path(pathName, filename)
    '/c/d/e/filename.ext'

    >>> pathName = os.path.join('/a','b','c','d1','e1','f1')
    >>> filename = os.path.join('/a','b','c','d2','e2','f2','filename.ext')
    >>> relative_path(pathName, filename)
    '../../../d2/e2/f2/filename.ext'

    '''

    if not path1:
        return path2
    common, (tail1, tail2) = common_prefix(path1.split(sep), path2.split(sep))
    if (not common or
        (len(common) == 1 and (common[0] == '' or ':' in common[0]))):
        return path2
    return sep.join([pardir]*len(tail1) + tail2)

def fix_path(path1, path2, maxpardir=2):
    '''
    >>> pathName = os.path.join('/a','b','c')
    >>> filename = os.path.join('/a','b','d','e','filename.ext')
    >>> fix_path(pathName, filename)
    '../d/e/filename.ext'

    >>> pathName = os.path.join('/a','b','c')
    >>> filename = os.path.join('/c','d','e','filename.ext')
    >>> fix_path(pathName, filename)
    '/c/d/e/filename.ext'

    >>> pathName = os.path.join('/a','b','c','d1','e1','f1')
    >>> filename = os.path.join('/a','b','c','d2','e2','f2','filename.ext')
    >>> fix_path(pathName, filename)
    '/a/b/c/d2/e2/f2/filename.ext'

    '''

    relpath = relative_path(path1, path2)
    if relpath.startswith(os.sep.join([os.pardir]*maxpardir)):
        return path2
    return relpath

def fromstring(str_):
    '''Try to convert a string in None, True or False, else simply remove
    quotes'''

    # @TODO: fix name
    if str_ == 'None':
        result = None
    elif str_.lower() in ('true', 'yes', 'on', 'y'):   #, '1'):
        result = True
    elif str_.lower() in ('false', 'no', 'off', 'n'):  #, '0'):
        result = False
    #~ for type_ in (int, long, float, complex):
        #~ try:
            #~ result = type_(result)
            #~ break
        #~ except ValueError:
            #~ pass
    else:
        result = remove_quote(str_)
    return result

def safe_eval(source, globals_=None, locals_=None):
    '''Evaluate the source in safe mode.

    >>> print safe_eval("('a','b','c',)")
    ('a', 'b', 'c')
    >>> print safe_eval("__import__('sys').exit(23)")
    Traceback (most recent call last):
      File "a.py", line 9, in ?
        print safe_eval("__import__('sys').exit(23)")
      File "a.py", line 4, in safe_eval
        raise ValueError, "Names %s not allowed" % (c.co_names,)
    ValueError: Names ('__import__', 'exit') not allowed

    '''

    c = compile(source, '<str>', 'eval')
    names = set(c.co_names)
    if globals_:
        names.difference_update(globals_)
    if locals_:
        names.difference_update(locals_)
    if names:
        raise ValueError, "Names %s not allowed" % (c.co_names,)
    return eval(c, globals_, locals_)

if sys.platform[:3] == 'win':
    def isexecutable(cmd):
        '''Check if "cmd" actually is an executable program.'''

        cmd = cmd.lower()
        if os.path.isfile(cmd) and cmd.endswith(('.exe', '.bat')):
            return True
        for ext in '.exe', '.bat':
            if os.path.isfile(cmd + ext):
                return True
        return False
else:
    def isexecutable(cmd):
        '''Check if "cmd" actually is an executable program.'''

        if os.path.isfile(cmd):
            mode = os.stat(cmd)[stat.ST_MODE]
            if ((mode & stat.S_IXUSR) or (mode & stat.S_IXGRP) or
                                                    (mode & stat.S_IXOTH)):
                return True
        return False

def which(cmd, env=None):
    '''Return the full path of the program (cnd) or None.

    >>> which('ls')
    '/bin/ls'

    '''

    if not env:
        env = os.environ

    for dir_ in env.get('PATH', '').split(os.pathsep):
        exe = os.path.join(dir_, cmd)
        if isexecutable(exe):
            return exe


if __name__ == '__main__':
    def _test(verbose=False):
        import doctest
        doctest.testmod(verbose=verbose)
    _test()
