### Copyright (C) 2006-2009 Antonio Valentino <a_valentino@users.sf.net>

### This file is part of BESTGUI.

### BESTGUI is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.

### BESTGUI is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.

### You should have received a copy of the GNU General Public License
### along with BESTGUI; if not, write to the Free Software
### Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA.

'''Viewers for images and parameters.'''

__author__  = 'Antonio Valentino <a_valentino@users.sf.net>'
__date__    = '$Date: 2009-04-13 10:09:18 +0200 (lun, 13 apr 2009) $'
__version__ = '$Revision: 704 $'

import os
import sys
import math
import shutil
import logging
import tempfile
import cStringIO
import ConfigParser

import gobject
import gtk

import utils
import gladewrap

class Viewer(object):
    '''Base class for BEST viewers'''

    def show_contents(self):
        pass

    def clear(self):
        pass

    def close(self):
        pass

class GladeViewer(Viewer, gladewrap.GladeObject):
    '''Base class for BEST viewers defined by a glade file'''

    def __init__(self, gladefile, root):
        #Viewer. __init__(self)
        gladewrap.GladeObject.__init__(self, gladefile, root)


class AnnotationsViewer(GladeViewer):
    '''Viewer for the BEST annotations dump files.'''

    def __init__(self, gladefile=None, root='annotations_scrolled'):
        if not gladefile:
            gladefile = utils.get_gladefile('bestgui.glade')
        super(AnnotationsViewer, self).__init__(gladefile, root)

        self.annotations_model = gtk.ListStore(str, str)
        self.annotations_model.set_sort_column_id(0, gtk.SORT_ASCENDING)

        self.annotations_view.set_model(self.annotations_model)
        self.annotations_view.set_search_column(0)

        for indx, name in enumerate(('Name', 'Value')):
            cell = gtk.CellRendererText()
            cell.set_property('editable', True)
            col = gtk.TreeViewColumn(name)
            col.pack_start(cell, expand=True)
            col.add_attribute(cell, 'text', indx)
            col.set_sort_column_id(indx)
            col.set_resizable(True)
            col.set_max_width(800)
            self.annotations_view.append_column(col)

        self.dumpfile = None

    def show_contents(self, filename=None):
        super(AnnotationsViewer, self).show_contents()

        if not filename:
            filename = self.dumpfile

        if filename:
            lines = open(filename).readlines()
            lines = [line.lstrip() for line in lines]
            fd = cStringIO.StringIO(''.join(lines))
            cfg = ConfigParser.SafeConfigParser()
            cfg.readfp(fd)
            for name, value in cfg.items('ANNOTATIONS'):
                #~ if len(value) > 256:
                    #~ logging.debug('very long value (%d characters) for "%s" '
                                  #~ 'not showed' % (len(value), name))
                    #~ value = 'VERY LONG LINE NOT SHOWED'
                self.annotations_model.append((name, value))
            # Set the new filename only if all is OK
            self.dumpfile = filename

    def clear(self):
        super(AnnotationsViewer, self).clear()
        self.annotations_model.clear()

class ESAAnnotationsViewer(GladeViewer):
    # @TODO: collapse/expand all
    #        improve search
    tags = (
        'Processing time',
        'Product type',
        'Sensor Mode',
        'Desired AP dataset',
        'Source',
        'Data format',
        'Facility id',
        'Format descriptor record',
        'File name',
        'Record name',
    )
    split_field_start = (0,  5, 48, 74,  91, 117)
    split_field_end = (4, 47, 73, 90, 116, 147)

    FIXEDSIZE = False

    def __init__(self, gladefile=None, root='annotations_scrolled'):

        if not gladefile:
            gladefile = utils.get_gladefile('bestgui.glade')
        super(ESAAnnotationsViewer, self). __init__(gladefile, root)

        self.annotations_model = gtk.TreeStore(str, str, str, str, str, str)

        self.annotations_view.set_model(self.annotations_model)
        if self.FIXEDSIZE:
            self.annotations_view.set_fixed_height_mode(True)
        self.annotations_view.set_search_column(1)
        self.annotations_view.set_headers_clickable(True) # @TODO: check
        #self.annotations_view.columns_autosize()
        try:
            # @COMPATYBILITY: set_grid_lines is new in PyGTK 2.10
            self.annotations_view.set_grid_lines(gtk.TREE_VIEW_GRID_LINES_BOTH)
        except AttributeError:
            pass

        for indx, (name, width) in enumerate((('Pos', 100),
                                              ('ESA field name', 300),
                                              ('Value', 200),
                                              ('Units', 100),
                                              ('Tag', 100),
                                              ('Remark', 200))):
            cell = gtk.CellRendererText()
            cell.set_property('editable', True)
            #cell.set_fixed_height_from_font(1)
            #cell.set_fixed_size(100, -1)
            #cell.width_chars = 15
            #cell.wrap_width = 150
            #cell.wrap_mode = gtk.WRAP_WORD

            col = gtk.TreeViewColumn(name)
            col.pack_start(cell, expand=True)
            col.add_attribute(cell, 'text', indx)
            col.set_resizable(True)
            #col.set_clickable(True)
            if self.FIXEDSIZE:
                col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
                col.set_min_width(width)
            self.annotations_view.append_column(col)

        self.annotations_file = None

    def show_contents(self, filename=None):
        super(ESAAnnotationsViewer, self).show_contents()

        if not filename:
            filename = self.annotations_file

        if filename:
            lines = open(filename).readlines()
            lines = lines[3:]
            model = self.annotations_model
            file_iter = None
            record_iter = None
            field = [[],[],[],[],[],[]]
            while lines:
                line = lines.pop(0)
                if '...: ' in line:
                    tag, value = line.split('...: ')
                    tag = tag.rstrip('.')
                    value = value.rstrip()
                    assert tag in self.tags
                    if tag  == 'File name':
                        file_iter = model.append(None,
                                                 ('', value,'', '', '', ''))
                    elif tag  == 'Record name':
                        assert file_iter is not None
                        record_iter = model.append(file_iter,
                                                   ('', value, '', '', '', ''))
                        assert lines
                        line = lines.pop(0)
                        assert line.startswith(' Pos Esa field name')
                elif line.startswith('-'):
                    # separator
                    if field != [[],[],[],[],[],[]]:
                        # put together the field components
                        for index, item in enumerate(field):
                            field[index] = ''.join(item).strip()
                        # add the field to the model
                        model.append(record_iter, field)
                        # reset the "field" variable
                        field = [[],[],[],[],[],[]]
                else:
                    # read the field items
                    for index, item in enumerate(field):
                        start = self.split_field_start[index]
                        end = self.split_field_end[index]
                        item.append(line[start:end])

            #~ self.annotations_view.expand_all()
            # expand first level nodes
            for row in model:
                self.annotations_view.expand_row(row.path, False)

            self.annotations_file = filename

    def clear(self):
        super(ESAAnnotationsViewer, self).clear()
        self.annotations_model.clear()

class ImageViewer(GladeViewer):
    '''Viewer for images'''

    # @TODO:
    #   - zoom selection
    #   - test with no image
    #   - allow rules hiding

    MINSIZE = 10
    MAXSIZE = 10000

    def __init__(self, gladefile=None, root='image_table'):
        if not gladefile:
            gladefile = utils.get_gladefile('bestgui.glade')
        super(ImageViewer, self).__init__(gladefile, root)

        self.imagefile = None
        self.data_pixbuf = None
        self.pixbuf = None
        self.pixbuf_offset = (0, 0)
        self.box_vertex = None
        self.box = None
        self.interp_type = gtk.gdk.INTERP_BILINEAR

        self.zoom_toolbar_separator.set_no_show_all(True)
        self.showgrid_togglebutton.set_no_show_all(True)
        self.image_preferences_button.set_no_show_all(True)

        self.hadjust = self.image_scrolled.get_hadjustment()
        self.vadjust = self.image_scrolled.get_vadjustment()
        self.hadjust.connect('value_changed', self.on_adjust_value_changed,
                             self.hruler, True)
        self.vadjust.connect('value_changed', self.on_adjust_value_changed,
                             self.vruler, False)

        self.default_zoom_in_ratio = 1.25
        self.default_zoom_out_ratio = 0.75

        self.zoom_entry = self.zoom_comboboxentry.get_child()
        self.zoom_entry.connect('activate', self.on_zoom_entry_activated)
        self.zoom_entry.connect('focus_out_event', self.on_focus_out_event)
        # @TODO: manually connect 'changed' signal and store the handler_id
        #        to call handler_block on it when needed

        self.zoom_comboboxentry.set_active(3)

    def show_contents(self, filename=None):
        super(ImageViewer, self).show_contents()
        if not filename:
            filename = self.imagefile

        if filename:
            logging.debug('gtk.gdk.pixbuf_new_from_file("%s") ...' % filename)
            self.data_pixbuf = gtk.gdk.pixbuf_new_from_file(filename)
            # @TODO: clean
            #~ try:
                #~ logging.debug('gtk.gdk.pixbuf_new_from_file("%s") ...' % filename)
                #~ self.data_pixbuf = gtk.gdk.pixbuf_new_from_file(filename)
            #~ except gobject.GError, e:
                #~ if 'file' in str(e):
                    #~ raise IOError(e.code, e.message, filename)
                #~ else:
                    #~ raise

            self.pixbuf = self.data_pixbuf
            width = self.pixbuf.get_width()
            height = self.pixbuf.get_height()
            self.hruler.set_range(0, width, 0, width)
            self.vruler.set_range(0, height, 0, height)
            self.box = None
            self.pixbuf_offset = (0, 0)
            self.drawingarea.set_size_request(width, height)
            self.center_image()
            if self.drawingarea.window:
                self.drawingarea.window.invalidate_rect(None, False)

            self.zoom_comboboxentry.set_active(3)
            self.zoom_comboboxentry.set_sensitive(True)
            # Set the new filename only if all is OK
            self.imagefile = filename
            return True

    def clear(self):
        super(ImageViewer, self).clear()
        #self.data_pixbuf = None # @TODO: clear or close?
        self.pixbuf = None
        self.zoom_comboboxentry.set_sensitive(False)
        self.zoom_comboboxentry.set_active(3)

    def zoom(self, ratio):
        self.zoom_entry.set_text('%d%%' % (100. * ratio))
        while gtk.events_pending():
            gtk.main_iteration()

    def _zoom(self, ratio):
        # ratio is the zoom ratio with respect the original image
        assert ratio > 0
        if not self.pixbuf:
            return
        current_width = self.data_pixbuf.get_width()
        current_height = self.data_pixbuf.get_height()
        width = int(current_width * ratio)
        height = int(current_height * ratio)
        if (width < self.MINSIZE or height < self.MINSIZE or
            width > self.MAXSIZE or height > self.MAXSIZE):
            ratio = self.get_current_ratio()
            self.zoom(ratio)
        else:
            if self.box:
                x, y, box_width, box_height = self.box
                # coordinates in the pixbug reference system
                x -= self.pixbuf_offset[0]
                y -= self.pixbuf_offset[1]
                # rescale factor with respect the current pixbuf (not
                # the data_pixbuf)
                scale = float(width) / self.pixbuf.get_width()
                # rescale the box and reset the pixbuf_offset
                self.box.x = x * scale + self.pixbuf_offset[0]
                self.box.y = y * scale + self.pixbuf_offset[1]
                self.box.width *= scale
                self.box.height *= scale

            self.pixbuf = self.data_pixbuf.scale_simple(width, height,
                                                        self.interp_type)
            self.drawingarea.set_size_request(width, height)
            self.center_image()
            #self.drawingarea.window.invalidate_rect(None, False)
            self.drawingarea.window.invalidate_rect(
                                    self.drawingarea.get_allocation(), True)

    def get_current_ratio(self):
        # @TODO: handle yratio !=xratio

        view_width = self.pixbuf.get_width()
        original_width = self.data_pixbuf.get_width()
        xratio = view_width / float(original_width)

        #view_height = self.pixbuf.get_height()
        #original_height = self.data_pixbuf.get_height()
        #yratio = view_height / float(original_height)

        return xratio #, yratio

    def get_available_size(self):
        '''Return the available size (widht, height) for the image'''
        border_width = self.image_viewport.get_border_width()
        border_width += 2 # empirical

        # @TODO: try to use drawingarea allocation
        rectangle = self.image_viewport.get_allocation()
        width = rectangle.width - border_width
        height = rectangle.height - border_width

        return int(width), int(height)

    def drawingarea_to_data_coor(self, x, y):
        '''Convert coordinates from the drawingarea reference system into
        the data_pixbuf one.'''
        ratio = self.get_current_ratio()
        x_data = (x - self.pixbuf_offset[0]) / ratio
        y_data = (y - self.pixbuf_offset[1]) / ratio

        return x_data, y_data

    def data_to_drawingarea_coor(self, x, y):
        '''Convert coordinates from the data_pixbuf reference system into
        the drawingarea one.'''
        ratio = self.get_current_ratio()
        x_area = x * ratio + self.pixbuf_offset[0]
        y_area = y * ratio + self.pixbuf_offset[1]

        return x_area, y_area

    def center_image(self):
        x, y, area_width, area_height = self.drawingarea.get_allocation()

        ratio = self.get_current_ratio()

        x = max(0, (area_width - self.pixbuf.get_width())/2)
        y = max(0, (area_height - self.pixbuf.get_height())/2)

        width = self.hruler.get_allocation().width / ratio
        height = self.vruler.get_allocation().height / ratio

        # set the rulers
        lower_limit, upper_limit, position, max_size = self.hruler.get_range()
        max_size = max(max_size, width)
        lower_limit = (-x + self.hadjust.value) / ratio
        upper_limit = lower_limit + width
        self.hruler.set_range(lower_limit, upper_limit, position, max_size)

        lower_limit, upper_limit, position, max_size = self.vruler.get_range()
        max_size = max(max_size, height)
        lower_limit = (-y + self.vadjust.value) / ratio
        upper_limit = lower_limit + height
        self.vruler.set_range(lower_limit, upper_limit, position, max_size)

        if self.box:
            ## remove the box
            #self.box = None

            #self.box_vertex = None # it should be already None
            old_offset = self.pixbuf_offset
            self.box.x += x - old_offset[0]
            self.box.y += y - old_offset[1]

        # set the new offset
        self.pixbuf_offset = (x,y)

    def center(self, xc, yc):
        adjust = self.hadjust
        size = adjust.page_size
        lower = max(xc - size/2, 0)
        upper = min(xc + size/2, self.pixbuf.get_width())
        adjust.clamp_page(lower, upper)

        adjust = self.vadjust
        size = adjust.page_size
        lower = max(yc - size/2, 0)
        upper = min(yc + size/2, self.pixbuf.get_height())
        adjust.clamp_page(lower, upper)

    def fit_ratio(self, width=None, height=None, mode='Fit'):
        if width is None:
            width = self.data_pixbuf.get_width()
        if height is None:
            height = self.data_pixbuf.get_height()

        widget_width, widget_height = self.get_available_size()

        # @TODO: needs some fix
        xratio = widget_width / float(width)
        if widget_height < xratio * height:
            # fix xratio taking into account the scrollbar size
            scrollbar = self.image_scrolled.get_vscrollbar()
            scrollbar_width = scrollbar.size_request()[0] + 5 # empirical
            xratio = (widget_width - scrollbar_width)/float(width)

        # @TODO: needs some fix
        yratio = widget_height / float(height)
        if widget_width < yratio * width:
            # fix yratio taking into account the scrollbar size
            scrollbar = self.image_scrolled.get_hscrollbar()
            scrollbar_height = scrollbar.size_request()[1] + 5 # empirical
            yratio = (widget_height - scrollbar_height)/float(height)

        if mode == 'Fit width':
            ratio = xratio
        elif mode == 'Fit height':
            ratio = yratio
        else:
            ratio = xratio
            if (ratio * height) > widget_height:
                ratio = yratio

        return ratio

    def on_drawingarea_expose_event(self, area, event=None):
        if self.pixbuf:
            x, y = self.pixbuf_offset
            area.window.draw_pixbuf(None, self.pixbuf, 0, 0, x, y)
            if self.box:
                box_gc = area.window.new_gc(function=gtk.gdk.INVERT)
                box_gc.line_width = 3
                area.window.draw_rectangle(box_gc, False, *self.box)
                #area.window.draw_rectangle(self.box_gc, False, *self.box)
        return True

    def on_table_size_allocate(self, widget, allocation):
        if self.pixbuf:
            self.center_image()
            if self.drawingarea.window:
                self.drawingarea.window.invalidate_rect(None, False)

    def on_adjust_value_changed(self, adjust, ruler, horiz):
        if horiz:
            span = self.image_scrolled.get_allocation().width
        else:
            span = self.image_scrolled.get_allocation().height
        lower_limit, upper_limit, position, max_size = ruler.get_range()
        ratio = self.get_current_ratio()
        lower_limit = adjust.value / ratio
        upper_limit = lower_limit + span / ratio
        ruler.set_range(lower_limit, upper_limit, position, max_size)
        self.drawingarea.window.invalidate_rect(None, False)
        while gtk.events_pending():
            gtk.main_iteration()

    def on_drawingarea_button_press_event(self, widget, event):
        if event.button == 1:
            if self.box:
                self.box = None
                self.drawingarea.window.invalidate_rect(None, False)
            self.box_vertex = (int(event.x), int(event.y))

    def on_drawingarea_button_release_event(self, widget, event):
        if self.pixbuf and event.button == 1:
            self.box_vertex = None
            if self.box and self.zoom_selection_button.get_active():
                xc = self.box.x + self.box.width/2.
                yc = self.box.y + self.box.height/2.
                xc, yc = self.drawingarea_to_data_coor(xc, yc)

                ratio = self.get_current_ratio()
                ratio *= self.fit_ratio(self.box.width, self.box.height)
                self.box = None
                self.zoom(ratio)

                xc, yc = self.data_to_drawingarea_coor(xc, yc)
                self.center(xc, yc)

    def on_drawingarea_motion_notify_event(self, widget, event, data=None):
        self.vruler.emit('motion_notify_event', event)
        self.hruler.emit('motion_notify_event', event)

        if event.is_hint:
            x, y, state = event.window.get_pointer()
        else:
            x = event.x
            y = event.y
            state = event.state

        if state & gtk.gdk.BUTTON1_MASK and self.box_vertex:
            x0 = self.box_vertex[0]
            y0 = self.box_vertex[1]

            width = abs(x0 - x)
            height = abs(y0 - y)

            x = min(int(x), x0)
            y = min(int(y), y0)

            self.box = gtk.gdk.Rectangle(x, y, width, height)
            self.drawingarea.window.invalidate_rect(None, False)

    def on_zoom_in(self, widget=None, data=None):
        if not self.pixbuf:
            return

        xc = self.hadjust.value + self.hadjust.page_size / 2.
        yc = self.vadjust.value + self.vadjust.page_size / 2.
        xc, yc = self.drawingarea_to_data_coor(xc, yc)

        ratio = self.get_current_ratio()
        ratio = math.ceil(100 * ratio * self.default_zoom_in_ratio) / 100.
        self.zoom(ratio)

        xc, yc = self.data_to_drawingarea_coor(xc, yc)
        self.center(xc, yc)

    def on_zoom_out(self, widget=None, data=None):
        if not self.pixbuf:
            return

        xc = self.hadjust.value + self.hadjust.page_size / 2.
        yc = self.vadjust.value + self.vadjust.page_size / 2.
        xc, yc = self.drawingarea_to_data_coor(xc, yc)

        ratio = self.get_current_ratio()
        ratio = math.floor(100 * ratio * self.default_zoom_out_ratio) / 100.
        self.zoom(ratio)

        xc, yc = self.data_to_drawingarea_coor(xc, yc)
        self.center(xc, yc)

    def _on_zoom_fit(self, widget=None, mode='Fit'):
        if not self.pixbuf:
            return

        image_width = self.data_pixbuf.get_width()
        image_height = self.data_pixbuf.get_height()
        widget_width, widget_height = self.get_available_size()

        if mode == 'Fit width':
            # @TODO: needs some fix
            xratio = widget_width / float(image_width)
            if widget_height < xratio * image_height:
                # fix xratio taking into account the scrollbar size
                scrollbar = self.image_scrolled.get_vscrollbar()
                scrollbar_width = scrollbar.size_request()[0] + 5 # empirical
                xratio = (widget_width - scrollbar_width)/float(image_width)
            ratio = xratio
        elif mode == 'Fit height':
            # @TODO: needs some fix
            yratio = widget_height / float(image_height)
            if widget_width < yratio * image_width:
                # fix yratio taking into account the scrollbar size
                scrollbar = self.image_scrolled.get_hscrollbar()
                scrollbar_height = scrollbar.size_request()[1] + 5 # empirical
                yratio = (widget_height - scrollbar_height)/float(image_height)
            ratio = yratio
        else:
            ratio = widget_width / float(image_width)       # xratio
            if (ratio * image_height) > widget_height:
                ratio = widget_height / float(image_height) # yratio

        self.zoom_entry.set_text('%d%%' % (100 * ratio))

    def on_zoom_fit(self, widget=None, mode='Fit'):
        if not self.pixbuf:
            return

        image_width = self.data_pixbuf.get_width()
        image_height = self.data_pixbuf.get_height()
        ratio = self.fit_ratio(image_width, image_height, mode)
        self.zoom(ratio)

    def on_zoom1(self, widget=None):
        if not self.pixbuf:
            return
        self.zoom(1.)

    def on_zoom_comboboxentry_changed(self, widget=None, data=None):
        if not self.pixbuf:
            return

        if self.zoom_entry.is_focus():
            return

        model = self.zoom_comboboxentry.get_model()
        index = self.zoom_comboboxentry.get_active()
        if index == -1:
            self.on_zoom_entry_activated(widget, data)
        else:
            value = model[index][0]
            if value in ('Fit width', 'Fit height', 'Fit'):
                self.on_zoom_fit(widget, value)
            else:
                ratio = float(value.rstrip(' %')) / 100.
                self._zoom(ratio)

    def on_zoom_entry_activated(self, widget=None, data=None):
        value = self.zoom_entry.get_text()
        try:
            ratio = float(value.rstrip(' %')) / 100.
            if '%' not in value:
                self.zoom_entry.set_text(value + '%')
        except ValueError, e:
            if value:
                logging.debug('invalid percentage "%s".' % value)
                percentage = round(self.get_current_ratio() * 100.)
                self.zoom_entry.set_text('%d%%' % percentage)
        else:
            self._zoom(ratio)

    def on_focus_out_event(self, event=None, data=None):
        self.zoom_entry.activate()

class QuickLookViewer(ImageViewer):
    def __init__(self, gladefile=None, root='image_table'):
        if not gladefile:
            gladefile = utils.get_gladefile('bestgui.glade')
        super(QuickLookViewer, self).__init__(gladefile, root)

        self.zoom_toolbar_separator.set_no_show_all(False)
        self.showgrid_togglebutton.set_no_show_all(False)
        self.zoom_toolbar_separator.show()
        self.showgrid_togglebutton.show()

        self.ql_imagefile = None
        self.qlg_imagefile = None

    def get_showgrid(self):
        return self.showgrid_togglebutton.get_active()

    def set_showgrid(self, value):
        if self.datadir is None:
            return

        self.showgrid_togglebutton.toggled()

    showgrid = property(get_showgrid, set_showgrid)

    def get_active_filename(self):
        if (self.ql_imagefile is None) and (self.qlg_imagefile is None):
            return

        if (self.ql_imagefile is None):
            filename = self.qlg_imagefile
            self.showgrid_togglebutton.set_sensitive(False)
        elif (self.qlg_imagefile is None):
            filename = self.ql_imagefile
            self.showgrid_togglebutton.set_sensitive(False)
        else:
            self.showgrid_togglebutton.set_sensitive(True)
            if self.showgrid:
                filename = self.qlg_imagefile
            else:
                filename = self.ql_imagefile
        return filename

    def show_contents(self, ql_imagefile=None, qlg_imagefile=None):
        if ql_imagefile is not None:
            self.ql_imagefile = ql_imagefile
            self.imagefile = None
        if qlg_imagefile is not None:
            self.qlg_imagefile = qlg_imagefile
            self.imagefile = None

        filename = self.get_active_filename()

        if filename:
            super(QuickLookViewer, self).show_contents(filename)

    def on_showgrid_toggled(self, widget=None, data=None):
        filename = self.get_active_filename()
        if not filename or filename == self.imagefile:
            return
        try:
            #~ value = self.zoom_entry.get_text()
            #~ box = self.box
            #~ # also set self.imagefile
            #~ ImageViewer.show_contents(self, imagefile)
            #~ self.zoom_entry.set_text(value)
            #~ self.box = box
            #~ self.drawingarea.window.invalidate_rectNone, False)
            width = self.pixbuf.get_width()
            height = self.pixbuf.get_height()
            self.data_pixbuf = gtk.gdk.pixbuf_new_from_file(filename)
            self.imagefile = filename
            self.pixbuf = self.data_pixbuf.scale_simple(width, height,
                                                        self.interp_type)
            #if self.drawingarea.window:
            self.drawingarea.window.invalidate_rect(None, False)

            w = self.data_pixbuf.get_width()
            h = self.data_pixbuf.get_height()
            new_aspectratio = float(w) / h
            old_aspectratio = float(width) / height
            if abs(old_aspectratio - new_aspectratio)/old_aspectratio > 0.01:
                logging.warn('the new aspect-raitio (%f) is different from'
                             'the old one (%f).' %
                                            (new_aspectratio, old_aspectratio))
        except (IOError, OSError), e:
            logging.warn(e)

class BaseBESTProductViewer(AnnotationsViewer, QuickLookViewer):
    def __init__(self, gladefile=None, root='hpaned'):
        if not gladefile:
            gladefile = utils.get_gladefile('bestgui.glade')
        super(BaseBESTProductViewer, self).__init__(gladefile, root)

    def show_contents(self, dumpfile=None,
                      ql_imagefile=None, qlg_imagefile=None):
        if dumpfile:
            self.dumpfile = dumpfile
        if ql_imagefile:
            self.ql_imagefile = ql_imagefile
            self.imagefile = None
        if qlg_imagefile:
            self.qlg_imagefile = qlg_imagefile
            self.imagefile = None

        super(BaseBESTProductViewer, self).show_contents()

class BESTProductViewer(BaseBESTProductViewer):
    def __init__(self, productfile, runfunc):
        super(BESTProductViewer, self).__init__()
        #self.image_preferences_button.show() # @TODO: complete

        self.product_file = productfile
        self.runfunc = runfunc
        self.datadir = None
        self._force_update = None
        self._run_count = 0

    def get_needs_update(self):
        if self._force_update is not None:
            return bool(self._force_update)

        try:
            files_ok = (os.access(self.dumpfile, os.R_OK) and
                        os.access(self.ql_imagefile, os.R_OK) and
                        os.access(self.qlg_imagefile, os.R_OK))
            return not files_ok
        except TypeError:
            return True

    def set_needs_update(self, value):
        '''Set the needs_update property

        True  -- force update
        False -- force no update
        None  -- auto update (the update is required only if one or both
                  data files do not exist

        '''

        self._force_update = value

    needs_update = property(get_needs_update, set_needs_update)

    # Result view interface
    def get_template(self):
        filename = utils.get_cfgfile('show_internal_product_template.ini')
        template = open(filename).read()
        return template

    def clear(self):
        super(BESTProductViewer, self).clear()
        self._run_count = 0

    def close(self):
        self.clear()
        if self.datadir:
            shutil.rmtree(self.datadir)
            self.datadir = None

    # @TODO: maybe a __del__ method could be usefull
    def __del__(self):
        #print 'BESTProductViewer.__del__'
        # if an exceptions is raised in the __init__ method the "datadir"
        # attribute could be not initialized
        if self.__dict__.get('datadir', None):
            def onerrorfunc():
                logging.warn('unable to remove the temporary folder',
                             exc_info=sys.exc_info())

            shutil.rmtree(self.datadir, onerror=onerrorfunc)
            self.datadir = None

    # ImageResultView methods
    def show_contents(self):
        if self.needs_update:
            self.generate()
        super(BESTProductViewer, self).show_contents()

    def generate(self, workdir=None, template=None):
        assert self.product_file

        if self._run_count > 0:
            raise RuntimeError('preview generation run more than once')

        # setup the temporary dirirectory for preview generation
        if self.datadir is None:
            # @TODO: improve name
            self.datadir = tempfile.mkdtemp('', 'tmp_preview_', workdir)
        elif not os.path.exists(self.datadir):
            os.mkdir(self.datadir)

        # set filenames
        # @TODO: set from template
        self.dumpfile = os.path.join(self.datadir, 'dump.txt')
        self.ql_imagefile = os.path.join(self.datadir, 'ql.tif')
        self.qlg_imagefile = os.path.join(self.datadir, 'qlg.tif')

        # setup the parameters file for BEST (ini file)
        if template is None:
            template = self.get_template()

        inputdir = os.path.dirname(self.product_file)
        filename = os.path.basename(self.product_file)
        outputdir = self.datadir # name required in string formatting
        # @TODO: insert here the parameters customization

        dirsep = os.sep

        inifilename = os.path.join(self.datadir, 'show_internal_product.ini')
        fd = open(inifilename, 'w')
        fd.write(template % locals())
        fd.close()

        # run best
        self._run_count = self._run_count + 1
        self.runfunc(inifilename)

if __name__ == '__main__':
    def test_annotations_viewer():
        mainwin = gtk.Window()
        mainwin.set_title('Annotation dump viewer')
        mainwin.resize(500, 400)
        mainwin.connect('destroy', gtk.main_quit)
        viewer = AnnotationsViewer()
        viewer.show_contents(dumpfile)
        mainwin.add(viewer.widget)
        mainwin.show_all()
        gtk.main()

    def test_esa_annotations_viewer():
        mainwin = gtk.Window()
        mainwin.set_title('ESA Annotations viewer')
        mainwin.resize(700, 600)
        mainwin.connect('destroy', gtk.main_quit)
        viewer = ESAAnnotationsViewer()
        viewer.show_contents(annotationsfile)
        mainwin.add(viewer.widget)
        mainwin.show_all()
        gtk.main()

    def test_image_viewer():
        mainwin = gtk.Window()
        mainwin.set_title('Image viewer')
        mainwin.resize(500, 700)
        mainwin.connect('destroy', gtk.main_quit)
        viewer = ImageViewer()
        viewer.show_contents(qlg_imagefile)
        mainwin.add(viewer.widget)
        mainwin.show_all()
        gtk.main()

    def test_quicklook_viewer():
        mainwin = gtk.Window()
        mainwin.set_title('Image viewer')
        mainwin.resize(500, 700)
        mainwin.connect('destroy', gtk.main_quit)
        viewer = QuickLookViewer()
        viewer.show_contents(imagefile, qlg_imagefile)
        mainwin.add(viewer.widget)
        mainwin.show_all()
        gtk.main()

    def test_product_viewer():
        mainwin = gtk.Window()
        mainwin.set_title('Product viewer')
        mainwin.resize(700, 600)
        mainwin.connect('destroy', gtk.main_quit)
        viewer = BaseBESTProductViewer()
        viewer.show_contents(dumpfile, imagefile, qlg_imagefile)
        mainwin.add(viewer.widget)
        mainwin.show_all()
        gtk.main()

    def test_best_product_viewer():
        mainwin = gtk.Window()
        mainwin.set_title('BEST product viewer')
        mainwin.resize(700, 600)
        mainwin.connect('destroy', gtk.main_quit)
        del BESTProductViewer.__del__
        viewer = BESTProductViewer(None, None)
        viewer.datadir = datadir
        viewer.dumpfile = dumpfile
        viewer.ql_imagefile = imagefile
        viewer.qlg_imagefile = qlg_imagefile
        viewer.show_contents()
        mainwin.add(viewer.widget)
        #mainwin.show_all()
        mainwin.show_all()
        gtk.main()

    #datadir = os.path.join(os.pardir, 'data', 'tmp_ers')
    datadir = os.path.join(os.pardir, 'data', 'tmp_envisat')
    dumpfile = os.path.join(datadir, 'dump.txt')
    annotationsfile = os.path.join(datadir, 'envisat-annotations.txt')
    imagefile = os.path.join(datadir, 'ql.tif')
    qlg_imagefile = os.path.join(datadir, 'qlg.tif')

    #~ test_annotations_viewer()
    #~ test_esa_annotations_viewer()
    #~ test_image_viewer()
    #~ test_quicklook_viewer()
    #~ test_product_viewer()
    #~ test_best_product_viewer()

    def test(filename):
        win = gtk.Window()
        win.resize(400,300)
        win.connect('destroy', gtk.main_quit)
        sw = gtk.ScrolledWindow()
        win.add(sw)
        img = gtk.Image()
        sw.add(img)

        img.set_from_file(filename)

        win.show_all()
        gtk.main()

    #~ datadir = r'C:\Documents and Settings\valentino\Documenti\tmp\tmp_preview_ctf1es'
    #~ imagefile = os.path.join(datadir, 'ql.tif')

    #~ datadir = r'C:\Documents and Settings\valentino\Documenti\tmp'
    #~ imagefile = os.path.join(datadir, 'quick_look_crop_grid.tif')

    datadir = r'C:\Documents and Settings\valentino\Documenti\tmp\libtiffpic'
    imagefile = os.path.join(datadir, 'cramps.tif')

    test(imagefile)

